import json
import logging.config
import os
from typing import Any, Optional

from cfgtools import Configuration
from helper import config_logger
from msgbus import (
    MBEasyContext,
    TK_SYS_ACK,
    TK_STORECFG_GET,
    FM_PARAM,
    TK_HV_GLOBALRESTART, MBMessage,
)
from systools import sys_parsefilepath
from ._BaseRepository import BaseRepository
from ._HvLogHandler import HvLogHandler
from ._PosModelRetriever import PosModelRetriever


def find_value_with_vars(
    cfg,                 # type: Configuration
    path,                # type: str
    default_value=None,  # type: Any
):
    # type: (...) -> Optional[str]
    value = cfg.find_value(path, default_value)
    if value is None:
        return value

    for variable_name in os.environ:
        variable_format = "{{{}}}".format(variable_name)
        if variable_format in value:
            value = value.replace(variable_format, os.environ[variable_name])
            if variable_name in ("BUNDLEDIR", "HVDATADIR"):
                value = value.replace("/./", os.sep).replace("//", os.sep).replace("/", os.sep)

    return value


def config_logger_parse_file_name(
    loader_path,              # type: str
    log_name='component',     # type: str
    configure_root=False,     # type: bool
    root_level=logging.WARN,  # type: int
    max_files=5,              # type: int
    max_size=5242880,         # type: int
    add_error_handler=False,  # type: bool
):
    # type: (...) -> None

    real_path = os.path.realpath(loader_path)
    if os.path.isfile(real_path) and real_path.endswith(".json"):
        with open(real_path, "rb") as f:
            json_loads = json.loads(f.read().decode("utf-8"))
            handlers = json_loads.get("handlers")
            if handlers:
                for handler in handlers:
                    file_name = sys_parsefilepath(handlers[handler].get("filename", ""))
                    if file_name:
                        handlers[handler]["filename"] = file_name

                loader_path = json_loads

    config_logger(loader_path, log_name, configure_root, root_level, max_files, max_size, add_error_handler)


def config_logger_with_hv_logger(
    loader_path,              # type: str
    log_name='component',     # type: str
    configure_root=True,      # type: bool
    root_level=logging.WARN,  # type: int
    max_files=5,              # type: int
    max_size=5242880,         # type: int
    add_error_handler=False,  # type: bool
):
    # type: (...) -> None

    config_logger(loader_path, log_name, configure_root, root_level, max_files, max_size, add_error_handler)

    if _is_old_log(loader_path):
        comp_logger = logging.getLogger(log_name)
        comp_logger.addHandler(HvLogHandler())

        if configure_root:
            root_logger = logging.getLogger()
            root_logger.addHandler(HvLogHandler())


def read_sw_config(
    mb_context,   # type: MBEasyContext
    key,          # type: str
):
    # type: (...) -> str

    reply = mb_context.MB_EasySendMessage("StoreWideConfig", token=TK_STORECFG_GET, format=FM_PARAM, data=key)
    if reply.token != TK_SYS_ACK:
        return None

    return str(reply.data)


def restart_system(
        mb_context,             # type: MBEasyContext
        event_type,             # type: str
        pos_id="",              # type: str
        menu_manager_user="",   # type: str
):
    # type: (...) -> MBMessage

    _notify_global_restart(event_type, mb_context, menu_manager_user, pos_id)

    return mb_context.MB_SendMessage(
        destination=mb_context.hv_service,
        token=TK_HV_GLOBALRESTART,
    )


def _notify_global_restart(
        event_type,         # type: str
        mb_context,         # type: MBEasyContext
        menu_manager_user,  # type: str
        pos_id              # type: str
):
    event_xml = '''<GLOBALRESTART posId="{}" userId="{}"/>'''
    data = event_xml.format(pos_id, menu_manager_user)
    mb_context.MB_EasyEvtSend(
        subject="GLOBAL_RESTART",
        type=event_type,
        xml=data,
        synchronous=True,
        timeout=-1,
    )


def _is_old_log(
    loader_path,  # type: str
):
    # type(...) -> bool

    real_path = os.path.realpath(loader_path)
    return not os.path.isfile(real_path) or not real_path.endswith(".json")



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Dict,
    Tuple,
)

from messagebus import MessageBus
from nreports.repository import (
    I18N,
    OrderAPIRepository,
    ProductRepository,
    StoreConfigRepository,
    UserRepository,
)
from nreports_api.shared import JSON


class DataBuilder(ABC):

    def __init__(
        self,
        message_bus: MessageBus,
        order_api_repository: OrderAPIRepository,
        store_config_repository: StoreConfigRepository,
        i18n_repository: I18N,
        images_path: Dict[str, str],
        constants: Dict[str, str],
        user_repository: UserRepository,
        product_repository: ProductRepository,
    ) -> None:

        self._message_bus = message_bus
        self.order_api_repository = order_api_repository
        self.store_config_repository = store_config_repository
        self.user_repository = user_repository
        self.i18n_repository = i18n_repository
        self.product_repository = product_repository
        self.images_path = images_path
        self.constants = constants

    @abstractmethod
    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        raise NotImplementedError()

    def get_builder_init(
        self,
    ) -> Tuple[
        MessageBus,
        OrderAPIRepository,
        StoreConfigRepository,
        I18N,
        Dict[str, str],
        Dict[str, str],
        UserRepository,
        ProductRepository,
    ]:

        return (
            self._message_bus,
            self.order_api_repository,
            self.store_config_repository,
            self.i18n_repository,
            self.images_path,
            self.constants,
            self.user_repository,
            self.product_repository,
        )



import json

from messageprocessorutil import UuidMessageProcessor
from ._AsyncResponseInteractor import AsyncResponseInteractor
from .dto import AsyncResponseRequest


class AsyncResponseProcessor(UuidMessageProcessor):
    def __init__(
        self,
        interactor: AsyncResponseInteractor,
    ):
        super().__init__()

        self.interactor = interactor

    def parse_data(self, event, data):
        data_dict = json.loads(data.decode("utf-8"))
        return AsyncResponseRequest(
            subject=event.subject,
            data=data_dict,
        )

    def call_business(self, request: AsyncResponseRequest):
        self.interactor.process_async_response(request=request)

    def format_response(
        self,
        message_bus,
        message,
        event,
        input_obj,
        result
    ):
        return

    def format_exception(
        self,
        message_bus,
        message,
        event,
        input_obj,
        exception
    ):
        return

    def format_parse_exception(
        self,
        message_bus,
        message,
        event,
        data,
        exception
    ):
        return



from datetime import datetime

from application.model import UpdateType, UpdateData, UpdateStep, UpdateStatus
from application.model.configuration import Configurations
from application.model.customexception import (
    ErrorInsertingNewUpdateVersion,
    ErrorUpdatingStep,
    ErrorUpdatingStatus,
)
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, DataType, TokenCreator
from mwhelper import (
    BaseRepository,
)
from persistence import Connection
from typing import Optional, Dict
from ._CommonPOSRepository import CommonPOSRepository

TK_RUPTURA_CLEAN_WITH_CATALOG_UPDATE = TokenCreator.create_token("0", "42", "11")


class CatalogUpdaterPOSRepository(BaseRepository):
    def __init__(self, message_bus, configs):
        # type: (MbContextMessageBus, Configurations) -> None

        super(CatalogUpdaterPOSRepository, self).__init__(message_bus.mbcontext)

        self.message_bus = message_bus
        self.configs = configs
        self.catalog_update_value = UpdateType.catalog.value
        self.catalog_json_update_value = UpdateType.catalog_json.value
        self.local_configs = configs.updaters[UpdateType.catalog.name]
        self.common_pos_repository = CommonPOSRepository(message_bus)

    def get_last_downloaded_version(self, type_id):
        # type: (str) -> Dict

        def inner_func(conn):
            # type: (Connection) -> Optional[int]

            update_value = self.catalog_update_value if type_id is None else type_id.value
            query = """SELECT UpdateId, UpdateName
                       FROM UpdatesController
                       WHERE TypeId = '{0}'
                        AND DownloadedDate IS NOT NULL
                        AND StatusId <> '{1}'
                       ORDER BY Id DESC
                       LIMIT 1""".format(update_value, UpdateStatus.error)

            cursor = conn.select(query)
            if not cursor.rows():
                return {}

            update_id = cursor.get_row(0).get_entry("UpdateId")
            update_name = cursor.get_row(0).get_entry("UpdateName")
            return {"updateId": update_id, "updateName": update_name}

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def has_catalog_pending_update_to_apply(self):
        # type: () -> Optional[UpdateData]

        def inner_func(conn):
            # type: (Connection) -> Optional[UpdateData]

            query = """
                SELECT
                    UpdateId,
                    UpdateName,
                    ObtainedDate,
                    DownloadedDate,
                    BackupDate,
                    AppliedDate,
                    NotifiedDate
                FROM UpdatesController
                    WHERE TypeId = '1' AND AppliedDate IS NULL
                    ORDER BY Id DESC
                LIMIT 1
            """

            cursor = conn.select(query)
            row = cursor.get_row(0)
            update_data = None

            if row:
                update_data = UpdateData()
                update_data.update_id = row.get_entry("UpdateId")
                update_data.update_name = row.get_entry("UpdateName")
                update_data.obtained_date = row.get_entry("ObtainedDate")
                update_data.downloaded_date = row.get_entry("DownloadedDate")
                update_data.backup_date = row.get_entry("BackupDate")
                update_data.applied_date = row.get_entry("AppliedDate")
                update_data.notified_date = row.get_entry("NotifiedDate")

            return update_data

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def has_already_applied_update(self):
        # type: () -> bool

        return self.common_pos_repository.has_already_applied_update(
            self.configs.persistence_name,
            self.catalog_update_value
        )

    def insert_new_update(self, update_id, update_name, type_id=None):
        # type: (str, str, UpdateType) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                update_value = self.catalog_update_value if type_id is None else type_id.value
                obtained_date = str(datetime.now().isoformat())
                conn.query("""INSERT OR REPLACE INTO UpdatesController
                              (UpdateId, UpdateName, TypeId, StatusId, ObtainedDate)
                              VALUES ('{0}', '{1}', '{2}', '{3}', '{4}')
                           """.format(update_id, update_name, update_value, UpdateStatus.pending, obtained_date))

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorInsertingNewUpdateVersion(ex)

    def skip_pending_updates(self, update_type):
        # type: (str) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                datetime_now = str(datetime.now().isoformat())
                conn.query(
                    """
                        UPDATE UpdatesController
                        SET StatusId = '{0}',
                        BackupDate = '{1}',
                        AppliedDate = '{1}',
                        NotifiedDate = '{1}'
                        WHERE AppliedDate IS NULL AND TypeId = {2}
                        """.format(UpdateStatus.skipped, datetime_now, update_type))

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorInsertingNewUpdateVersion(ex)

    def get_pending_update_to_notify(self):
        # type: () -> Optional[str]

        def inner_func(conn):
            # type: (Connection) -> Optional[str]

            query = """
                        SELECT Id, UpdateId
                        FROM UpdatesController
                        WHERE TypeId = 1 AND AppliedDate IS NOT NULL AND NotifiedDate is NULL
                        ORDER BY Id DESC LIMIT 1
                    """

            cursor = conn.select(query)
            if not cursor.rows():
                return None, None

            database_id = cursor.get_row(0).get_entry("Id")
            update_id = cursor.get_row(0).get_entry("UpdateId")
            return database_id, update_id

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def update_step(self, update_id, update_step, type_id=None):
        # type: (int, UpdateStep, UpdateType) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                update_value = self.catalog_update_value if type_id is None else type_id.value
                datetime_now = str(datetime.now().isoformat())

                if update_step == UpdateStep.all:
                    conn.query(
                        """UPDATE UpdatesController
                               SET ObtainedDate = IFNULL(ObtainedDate, '{0}'),
                                   DownloadedDate = IFNULL(DownloadedDate, '{0}'),
                                   BackupDate = IFNULL(BackupDate, '{0}'),
                                   AppliedDate = IFNULL(AppliedDate, '{0}'),
                                   NotifiedDate = IFNULL(NotifiedDate, '{0}')
                               WHERE UpdateId = '{1}' AND TypeId = '{2}'
                            """.format(datetime_now, update_id, update_value)
                    )
                elif update_step == UpdateStep.notify:
                    conn.query(
                        """
                                UPDATE UpdatesController
                                SET '{0}' = '{1}'
                                WHERE
                                    TypeId = '{2}' AND
                                    NotifiedDate IS NULL
                            """.format(update_step, datetime_now, update_value)
                    )
                else:
                    conn.query(
                        """UPDATE UpdatesController
                               SET '{0}' = '{1}'
                               WHERE UpdateId = '{2}' AND TypeId = '{3}'
                            """.format(update_step, datetime_now, update_id, update_value)
                    )

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorUpdatingStep(ex)

    def update_status(self, update_id, update_status):
        # type: (int, UpdateStatus) -> None
        try:
            def inner_func(conn):
                # type: (Connection) -> None

                conn.query(
                    """
                        UPDATE UpdatesController
                        SET
                            StatusId = '{0}'
                        WHERE
                            UpdateId = '{1}' AND
                            TypeId = '{2}'
                    """.format(update_status, update_id, self.catalog_update_value)
                )

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorUpdatingStatus(ex)



import ast
import json
import textwrap
from pathlib import Path


def iter_namespace_pkgs(namespace):
    parts = namespace.split(".")
    for i in range(len(parts)):
        yield ".".join(parts[: i + 1])


def build_namespace_package(tmpdir, name, version="1.0", impl="pkg_resources"):
    src_dir = tmpdir / name
    src_dir.mkdir()
    setup_py = src_dir / 'setup.py'
    namespace, _, rest = name.rpartition('.')
    namespaces = list(iter_namespace_pkgs(namespace))
    setup_args = {
        "name": name,
        "version": version,
        "packages": namespaces,
    }

    if impl == "pkg_resources":
        tmpl = '__import__("pkg_resources").declare_namespace(__name__)'
        setup_args["namespace_packages"] = namespaces
    elif impl == "pkgutil":
        tmpl = '__path__ = __import__("pkgutil").extend_path(__path__, __name__)'
    else:
        raise ValueError(f"Cannot recognise {impl=} when creating namespaces")

    args = json.dumps(setup_args, indent=4)
    assert ast.literal_eval(args)  # ensure it is valid Python

    script = textwrap.dedent(
        """\
        import setuptools
        args = {args}
        setuptools.setup(**args)
        """
    ).format(args=args)
    setup_py.write_text(script, encoding='utf-8')

    ns_pkg_dir = Path(src_dir, namespace.replace(".", "/"))
    ns_pkg_dir.mkdir(parents=True)

    for ns in namespaces:
        pkg_init = src_dir / ns.replace(".", "/") / '__init__.py'
        pkg_init.write_text(tmpl, encoding='utf-8')

    pkg_mod = ns_pkg_dir / (rest + '.py')
    some_functionality = 'name = {rest!r}'.format(**locals())
    pkg_mod.write_text(some_functionality, encoding='utf-8')
    return src_dir


def build_pep420_namespace_package(tmpdir, name):
    src_dir = tmpdir / name
    src_dir.mkdir()
    pyproject = src_dir / "pyproject.toml"
    namespace, _, rest = name.rpartition(".")
    script = f"""\
        [build-system]
        requires = ["setuptools"]
        build-backend = "setuptools.build_meta"

        [project]
        name = "{name}"
        version = "3.14159"
        """
    pyproject.write_text(textwrap.dedent(script), encoding='utf-8')
    ns_pkg_dir = Path(src_dir, namespace.replace(".", "/"))
    ns_pkg_dir.mkdir(parents=True)
    pkg_mod = ns_pkg_dir / (rest + ".py")
    some_functionality = f"name = {rest!r}"
    pkg_mod.write_text(some_functionality, encoding='utf-8')
    return src_dir


def make_site_dir(target):
    """
    Add a sitecustomize.py module in target to cause
    target to be added to site dirs such that .pth files
    are processed there.
    """
    sc = target / 'sitecustomize.py'
    target_str = str(target)
    tmpl = '__import__("site").addsitedir({target_str!r})'
    sc.write_text(tmpl.format(**locals()), encoding='utf-8')



import pyscripts
import sysactions

from bustoken import TK_SLCTRL_PMGR_GET_MODIFIERS
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
)
from posot import OrderTakerException


mb_context = pyscripts.mbcontext
_modifiers = None


@sysactions.action
def get_modifiers_data(_):

    global _modifiers

    if _modifiers is None:
        message_response = mb_context.MB_EasySendMessage(
            dest_name='ORDERMGR',
            token=TK_SLCTRL_PMGR_GET_MODIFIERS,
            format=FM_PARAM,
        )
        if message_response.token != TK_SYS_ACK:
            raise OrderTakerException(
                code=-1,
                descr=message_response.data,
            )
        _modifiers = message_response.data

    return _modifiers



from logging import Logger
from production.box import ProductionBox
from typing import List, Union

from ._OrderChangerProductionBox import OrderChangerProductionBox


class SkipCourseBox(OrderChangerProductionBox):
    def __init__(self, name, sons, skip_pod_types, skip_sale_types, logger=None):
        # type: (str, Union[ProductionBox, List[ProductionBox]], List[str], List[str], Logger) -> None  # noqa
        super(SkipCourseBox, self).__init__(name, sons, logger)
        self.skip_pod_types = skip_pod_types
        self.skip_sale_types = skip_sale_types

    def change_order(self, order):
        order.skip_course = self.set_skip_course(order)
        return order

    def set_skip_course(self, order):
        return self.skip_course_by_pod_type(order) or self.skip_course_by_sale_type(order)

    def skip_course_by_pod_type(self, order):
        return order.pod_type in self.skip_pod_types

    def skip_course_by_sale_type(self, order):
        return order.sale_type in self.skip_sale_types



from systools import sys_log_exception

from actions.config import must_print_delivery_coupon
from actions.printer import send_json_data_print
from api.nreports import get_nreports_service
from model.customexception import DeliveryCouponPrintError
from typing import Optional
from utilfunctions import get_custom_property_value


def print_delivery_coupon(
    pos_id,         # type: str
    order_id,       # type: str
):
    # type: (...) -> None

    if not must_print_delivery_coupon():
        return

    brand = get_custom_property_value(
        pos_id=pos_id,
        order_id=order_id,
        custom_property='BRAND'
    )

    _print_delivery_coupon_text(
        pos_id=pos_id,
        order_id=order_id,
        brand=brand,
    )


def reprint_delivery_coupons(
    pos_id,         # type: str
    order_id,       # type: str
):
    # type: (...) -> None

    _print_delivery_coupon_text(
        pos_id=pos_id,
        order_id=order_id,
    )


def _print_delivery_coupon_text(
    pos_id,         # type: str
    order_id,       # type: str
    brand=None,     # type: Optional[str]
):
    # type: (...) -> None

    try:
        print_data = get_nreports_service().delivery_coupon(
            pos_id=pos_id,
            order_id=order_id,
        )
        send_json_data_print(
            pos_id=pos_id,
            print_data=print_data,
            brand=brand,
        )
    except Exception:
        msg = "Error printing Delivery Coupon"
        sys_log_exception(message=msg)
        raise DeliveryCouponPrintError(msg)



from typing import Dict, Any
from pos_model import Order

from old_helper import validar_cnpj, validar_cpf

from _SatXmlPartBuilder import SatXmlPartBuilder
from _ContextKeys import ContextKeys


class DestinatarioBuilder(SatXmlPartBuilder):
    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        # Dados Destinatario (cliente)
        xml = ""
        cpf_cnpj = order.customer_cpf
        context[ContextKeys.cpf_cnpj] = cpf_cnpj

        if not cpf_cnpj:
            xml += "<dest/>"
        else:
            if len(cpf_cnpj) == 11 and validar_cpf(cpf_cnpj):
                xml += "<dest><CPF>{}</CPF></dest>".format(cpf_cnpj)
            elif len(cpf_cnpj) == 14 and validar_cnpj(cpf_cnpj):
                xml += "<dest><CNPJ>{}</CNPJ></dest>".format(cpf_cnpj)
            else:
                xml += "<dest/>"

        return xml



from messagebus import (
    TokenCreator,
    TokenPriority,
)

TK_CMP_MESSAGE = TokenCreator.create_token(
    priority=TokenPriority.high,
    message_group="3",
    number="6",
)


TABLE_MGR_MESSAGE_GROUP = "C2"

TK_TABLE_MGR_GET_SECTORS = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=TABLE_MGR_MESSAGE_GROUP,
    number="1",
)


TK_TABLE_MGR_START_SERVICE = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=TABLE_MGR_MESSAGE_GROUP,
    number="2",
)

TK_TABLE_MGR_TABLE_SERVICE_PICTURE = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=TABLE_MGR_MESSAGE_GROUP,
    number="3",
)

TK_TABLE_MGR_START_TAB = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=TABLE_MGR_MESSAGE_GROUP,
    number="4",
)

TK_TABLE_MGR_TAB_SERVICE_PICTURE = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=TABLE_MGR_MESSAGE_GROUP,
    number="5",
)

CREATE_TAB_FROM_SPLIT_SERVICE = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=TABLE_MGR_MESSAGE_GROUP,
    number="6",
)



# -*- coding: utf-8 -*-

import logging
from xml.etree import cElementTree as eTree

from api.order import get_order_service
from custom_sysactions import block_action_if_pos_is_blocked
from peripherals.scale import PosScaleEvents, ScaleEvent, WeightComparer
from sysactions import (
    StopAction,
    action,
    get_posot,
    show_messagebox,
    show_keyboard,
    close_asynch_dialog,
)
from thread_local import get_current_model
from typing import Optional
from utilfunctions import get_device_service

from .drive_thru import validate_can_change_order
from ..products.get_products_with_measure_unit import is_a_product_with_measure_unit

waiting_scale = False
logger = logging.getLogger("PosActions")


@action
def doCompleteOption(
    pos_id,                 # type: str
    context,                # type: str
    part_code,              # type: str
    qty="1",                # type: str
    line_number="",         # type: str
    sale_type="EAT_IN",     # type: str
    subst="",               # type: str
    seat="",                # type: str
    comment=None,           # type: Optional[str]
    barcode_number=None,    # type: Optional[str]
):
    # type: (...) -> Optional[str]

    from posactions import (
        is_table_service,
        is_changing_order,
    )

    global waiting_scale

    model = get_current_model()

    order_service = get_order_service(pos_id=pos_id)
    order_picture = order_service.get_order_picture()

    is_order_table_service = is_table_service(model=model, order_picture=order_picture)
    if is_order_table_service and is_changing_order(model=model, order_picture=order_picture):
        show_messagebox(
            posid=pos_id,
            message="$CANNOT_ADD_ITEMS_IN_AN_OPENED_ORDER",
            icon="error",
        )
        raise StopAction()

    order_is_in_progress = order_picture and order_picture.get("state") == "IN_PROGRESS"
    if not is_order_table_service or not order_is_in_progress:
        _verify_pos_is_blocked(pos_id=pos_id, model=model)

    if order_is_in_progress:
        validate_can_change_order(
            pos_id=pos_id,
            order_picture=order_picture,
            line_to_replace=subst,
            context=context,
            line_number=line_number,
            sale_type=sale_type,
        )

    if is_a_product_with_measure_unit(product_code=part_code) and not isinstance(qty, float):
        if waiting_scale:
            return

        try:
            waiting_scale = True
            qty = _get_weight_qty(pos_id=pos_id)
        finally:
            waiting_scale = False

    sale_type = sale_type or "EAT_IN"
    return _do_complete_option(
        pos_id=pos_id,
        model=model,
        subst=subst,
        line_number=line_number,
        context=context,
        part_code=part_code,
        sale_type=sale_type,
        comment=comment,
        qty=qty,
        seat=seat,
        barcode_number=barcode_number,
        order_picture=order_picture,
    )


def _do_complete_option(
    pos_id,             # type: str
    model,              # type: eTree.Element
    subst,              # type: str
    line_number,        # type: str
    context,            # type: str
    part_code,          # type: str
    sale_type,          # type: str
    comment,            # type: str
    qty,                # type: str
    seat,               # type: str
    barcode_number,     # type: str
    order_picture,      # type: eTree.Element
):
    # type: (...) -> Optional[str]

    from posactions import (
        doClearOptionItem,
        doOption,
    )

    sale_xml = None
    pos_ot = get_posot(model=model)
    try:
        if subst:
            if qty != "0":
                pos_ot.blkopnotify = True

            doClearOptionItem(
                pos_id=pos_id,
                linenumber=line_number,
                item_id=subst,
                pos_ot=pos_ot,
            )
            pos_ot.blkopnotify = False

        if qty != "0":
            sale_xml = doOption(
                pos_id=pos_id,
                context=context,
                part_code=part_code,
                qty=qty,
                line_number=line_number,
                sale_type=sale_type,
                comment=comment,
                pos_ot=pos_ot,
                seat=seat,
                barcode_number=barcode_number,
                order_picture=order_picture,
            )

        return eTree.tostring(sale_xml).decode("utf-8") if sale_xml is not None else None
    finally:
        pos_ot.blkopnotify = False


@block_action_if_pos_is_blocked
def _verify_pos_is_blocked(
    pos_id,     # type: str
    model,      # type: eTree.Element
):
    # type: (...) -> None

    return


def _get_weight_qty(
    pos_id,    # type: str
):
    # type: (...) -> str

    weight_qty = _get_scale_weight(pos_id=pos_id)
    if weight_qty is None:
        weight_qty = show_keyboard(
            posid=pos_id,
            message="$TYPE_WEIGHT_QTY",
            title="",
            mask="weight",
            defvalue="0.000",
            numpad=True,
        )
    if weight_qty in (None, "", "."):
        raise StopAction()

    if weight_qty == "0":
        raise StopAction()

    return str(float(weight_qty))


def _get_scale_weight(
    pos_id,    # type: str
):
    # type: (...) -> str

    scale_name = get_device_service(
        pos_id=pos_id,
        device_type="scale",
    )
    weight = _scale_wait_data(
        pos_id=pos_id,
        service_name=scale_name,
    )
    return weight


def _scale_wait_data(
    pos_id,          # type: str
    service_name,    # type: str
    timeout=2000,    # type: Optional[float]
):
    # type: (...) -> Optional[str]

    scale_event = PosScaleEvents().create_scale_event(
        pos_id=pos_id,
        service_name=service_name,
    )

    return _get_scaled_data_weight(
        pos_id=pos_id,
        service_name=service_name,
        scale_event=scale_event,
        timeout=timeout,
    )


def _get_scaled_data_weight(
    pos_id,          # type: str
    service_name,    # type: str
    scale_event,     # type: ScaleEvent
    timeout,         # type: float
):
    # type: (...) -> Optional[str]

    from posactions import pos_scale_dialogs

    dialog_id = None
    weight_comparer = WeightComparer()
    try:
        while True:
            if dialog_id is None:
                dialog_id = show_messagebox(
                    posid=pos_id,
                    title="$WAIT_SCALE_WEIGHT",
                    message="$WAIT_SCALE_WEIGHT",
                    buttons="$CANCEL",
                    asynch=True,
                )
                pos_scale_dialogs.append(str(dialog_id))
            elif str(dialog_id) not in pos_scale_dialogs:
                return ""

            scale_event.wait(float(timeout / 1000.0))
            if scale_event.weight is None:
                return None

            if scale_event.weight > 0 and weight_comparer.is_weight_equal_to_previous(weight=scale_event.weight):
                close_asynch_dialog(
                    posid=pos_id,
                    dlgid=dialog_id,
                )
                dialog_id = None

                selected_option = show_messagebox(
                    posid=pos_id,
                    message="$CONFIRM_SCALE_WEIGHT|{0}".format("%.3f" % scale_event.weight),
                    buttons="$OK|$GET_WEIGHT_AGAIN|$CANCEL",
                )

                if selected_option == 0:
                    return scale_event.weight
                elif selected_option == 2:
                    return ""
    finally:
        _update_scale_event(
            pos_id=pos_id,
            dialog_id=dialog_id,
            scale_event=scale_event,
        )


def _update_scale_event(
    pos_id,         # type: str
    dialog_id,      # type: str
    scale_event,    # type: ScaleEvent
):
    # type: (...) -> None

    if pos_id in PosScaleEvents().get_wait_pos_id_event_dict():
        PosScaleEvents().finish_event(scale_event=scale_event)

    if dialog_id is not None:
        close_asynch_dialog(
            posid=pos_id,
            dlgid=dialog_id,
        )



from old_helper import round_half_away_from_zero
from pos_model import Order
from typing import (
    Any,
    Dict,
)

from _ContextKeys import ContextKeys
from _SatXmlPartBuilder import SatXmlPartBuilder
from satbuilder import (
    DetBuilder,
    DetFeeBuilder,
    DetTipBuilder,
)


class DetListBuilder(SatXmlPartBuilder):

    def __init__(
        self,               # type: DetListBuilder
        det_builder,        # type: DetBuilder
        det_tip_builder,    # type: DetTipBuilder
        det_fee_builder,    # type: DetFeeBuilder
    ):
        # type: (...) -> None

        self.det_builder = det_builder
        self.det_tip_builder = det_tip_builder
        self.det_fee_builder = det_fee_builder

    def build_xml(
        self,       # type: DetListBuilder
        order,      # type: Order
        context,    # type: Dict[unicode, Any]
    ):
        # type: (...) -> str

        xml = ""

        context[ContextKeys.total_vbc] = 0
        context[ContextKeys.total_prod] = 0
        context[ContextKeys.total_icms] = 0
        context[ContextKeys.total_pis] = 0
        context[ContextKeys.total_cofins] = 0
        context[ContextKeys.item_no] = 1

        fee_items = self.det_fee_builder.fee_service.get_fee_sale_items(order=order)

        for sale_item in order.sale_items:
            if fee_items:
                sale_item.item_price = round_half_away_from_zero(
                    number=sale_item.item_price - sale_item.fee_included,
                    precision=3,
                )
            context[ContextKeys.current_sale_item] = sale_item
            xml += self.det_builder.build_xml(
                order=order,
                context=context,
            )
            context[ContextKeys.item_no] += 1

        xml += self.det_fee_builder.build_xml(
            order=order,
            context=context,
        )
        xml += self.det_tip_builder.build_xml(
            order=order,
            context=context,
        )

        return xml



# coding:utf-8

import random


def random_jitter(value: float) -> float:
    """Jitter the value a random number of milliseconds.

    This adds up to 1 second of additional time to the original value.
    Prior to backoff version 1.2 this was the default jitter behavior.

    Args:
        value: The unadulterated backoff value.
    """
    return value + random.random()


def full_jitter(value: float) -> float:
    """Jitter the value across the full range (0 to value).

    This corresponds to the "Full Jitter" algorithm specified in the
    AWS blog's post on the performance of various jitter algorithms.
    (http://www.awsarchitectureblog.com/2015/03/backoff.html)

    Args:
        value: The unadulterated backoff value.
    """
    return random.uniform(0, value)



# -*- coding: utf-8 -*-
import iso8601
import json
import uuid

from xml.etree import cElementTree as eTree
from old_helper import convert_from_localtime_to_utc
from typing import Dict, List

from api.order import get_order_service
from order_api.dto import CustomProperty


def create_delivery_manual_order_json(
    pos_id,     # type: int
    xml_order,  # type: eTree.Element
):
    # type: (...) -> None

    name = _get_custom_property(
        xml_order=xml_order,
        custom_property="CUSTOMER_NAME",
    )
    phone = _get_custom_property(
        xml_order=xml_order,
        custom_property="CUSTOMER_PHONE",
    )
    document = _get_custom_property(
        xml_order=xml_order,
        custom_property="CUSTOMER_DOC",
    )
    state = _get_custom_property(
        xml_order=xml_order,
        custom_property="STATE",
    )
    city = _get_custom_property(
        xml_order=xml_order,
        custom_property="CITY",
    )
    neighborhood = _get_custom_property(
        xml_order=xml_order,
        custom_property="NEIGHBORHOOD",
    )
    reference = _get_custom_property(
        xml_order=xml_order,
        custom_property="ADDRESS_REFERENCE",
    )
    complement = _get_custom_property(
        xml_order=xml_order,
        custom_property="COMPLEMENT",
    )
    zip_code = _get_custom_property(
        xml_order=xml_order,
        custom_property="POSTAL_CODE",
    )
    number = _get_custom_property(
        xml_order=xml_order,
        custom_property="STREET_NUMBER",
    )
    street = _get_custom_property(
        xml_order=xml_order,
        custom_property="STREET_NAME",
    )
    need_logistics = _get_custom_property(
        xml_order=xml_order,
        custom_property="NEED_LOGISTICS",
    )
    pickup_type = _get_custom_property(
        xml_order=xml_order,
        custom_property="PICKUP_TYPE",
    )

    formatted_address = "{0}, {1} - {2} - {3}".format(street, number, neighborhood, city) if street and number else ""

    all_sale_lines = xml_order.findall("SaleLine")
    main_sale_lines = [x for x in all_sale_lines if x.get("level") == "0"]

    create_at = iso8601.parse_date(xml_order.get("createdAt", ""))
    create_at = convert_from_localtime_to_utc(localtime=create_at).strftime("%Y-%m-%d %H:%M:%S")
    order_id = xml_order.get("orderId", "-")
    remote_order_id = str(uuid.uuid4())
    manual_delivery_order_json = {
        "id": remote_order_id,
        "items": _get_items(
            all_sale_lines=all_sale_lines,
            sale_lines=main_sale_lines,
        ),
        "custom_properties": [
            {
                "key": "contact_number",
                "value": phone
            }, {
                "key": "customer_doc",
                "value": document
            }, {
                "key": "customer_name",
                "value": name
            }, {
                "key": "customer_phone",
                "value": phone
            }, {
                "key": "partner",
                "value": 'MANUAL'
            }, {
                "key": "created_at",
                "value": create_at
            }, {
                "key": "short_reference",
                "value": order_id
            }, {
                "key": "formatted_address",
                "value": formatted_address
            }, {
                "key": "street_name",
                "value": street
            }, {
                "key": "postal_code",
                "value": zip_code
            }, {
                "key": "need_logistics",
                "value": need_logistics
            }, {
                "key": "delivery_method",
                "value": 'DEFAULT'
            }, {
                "key": "logistic_integration_status",
                "value": 'Waiting'
            }, {
                "key": "pickup_type",
                "value": 'delivery'
            }, {
                "key": "manual_order_id",
                "value": remote_order_id
            },
        ],
        "pickup": {
            "type": pickup_type,
            "address": {
                "state": state,
                "city": city,
                "neighborhood": neighborhood,
                "reference": reference,
                "complement": complement,
                "postalCode": zip_code,
                "formattedAddress": formatted_address,
                "streetNumber": number,
                "streetName": street
            }
        },
        "totalPrice": float(xml_order.get("totalGross", 0)),
        "partner": "Delivery",
        "shortReference": xml_order.get("orderId", "-"),
        "subTotal": float(xml_order.get("totalGross", 0)),
        "createAt": xml_order.get("createdAt", "") + "Z"
    }

    all_custom_properties = _parse_custom_properties(parsed_json=manual_delivery_order_json)

    all_custom_properties.extend([
        CustomProperty(
            key="MANUAL_DELIVERY_ORDER_JSON",
            value=json.dumps(manual_delivery_order_json),
        ),
        CustomProperty(
            key="INTEGRATED_MANUAL_DELIVERY",
            value="TRUE",
        ),
        CustomProperty(
            key="REMOTE_ORDER_ID",
            value=remote_order_id,
        ),
    ])
    get_order_service(pos_id=pos_id).set_order_custom_properties(custom_properties=all_custom_properties)


def _parse_custom_properties(
    parsed_json,   # type: Dict[str, str]
):
    # type: (...) -> List[CustomProperty]

    all_custom_properties = []
    json_custom_properties = parsed_json.get("custom_properties") or []

    for json_custom_property in json_custom_properties:
        if json_custom_property["value"]:
            all_custom_properties.append(CustomProperty(
                key=json_custom_property["key"].upper(),
                value=json_custom_property["value"],
            ))

    return all_custom_properties


def _get_custom_property(
    xml_order,              # type: eTree
    custom_property,        # type: str
):
    # type: (...) -> str

    value = ""
    current_property = xml_order.find("CustomOrderProperties/OrderProperty[@key='{}']".format(custom_property))
    if current_property is not None:
        value = current_property.get("value", "").encode("utf-8")

    return value


def _get_items(
    all_sale_lines,     # type: List
    sale_lines,         # type: List
    level=0,            # type: int
):
    # type: (...) -> List

    items = []
    for sale_line in sale_lines:
        quantity = float(sale_line.get("multipliedQty", 0))
        if quantity == 0:
            continue

        line_number = sale_line.get("lineNumber")
        unit_price = float(sale_line.get("unitPrice", 0))
        overwritten_unit_price = float(sale_line.get("overwrittenUnitPrice", 0))
        price = (overwritten_unit_price or unit_price) * (quantity or 1)
        son_level = str(level + 1)
        son_item_id = sale_line.get("itemId") + "." + sale_line.get("partCode")
        child_sale_lines = [
            x for x in all_sale_lines
            if x.get("lineNumber") == line_number and x.get("level") == son_level and x.get("itemId") == son_item_id
        ]
        items.append(
            {
                "price": price,
                "parts": _get_items(
                    all_sale_lines,
                    child_sale_lines,
                    level + 1,
                ),
                "partCode": sale_line.get("partCode", ""),
                "observation": "",
                "quantity": quantity if sale_line.get("itemType", "") != "OPTION" else 1,
                "itemType": sale_line.get("itemType", "")
            }
        )

    return items




class ProdStates(object):

    NORMAL = 'NORMAL'        #: Order is in progress
    SERVED = 'SERVED'        #: Order has been assembled
    HELD = 'HELD'            #: Order has stopped being prepared
    READY = 'READY'          #: Order is ready for the customer to pick up
    DELIVERED = 'DELIVERED'  #: Order has been delivered to the customer
    INVALID = 'INVALID'      #: Order is invalid and should be ignored

    # List of all known production order states
    ALL = tuple((name for name in dir() if not name.startswith("_")))

    def __init__(self):
        return



import json
from typing import (
    Any,
    Dict,
)

from nreports.builders import DataBuilder
from nreports.utils import (
    convert_from_utf_to_localtime,
    format_customer_doc,
    get_logo_image_str,
)
from nreports_api.shared import JSON
from ._DeliveryOrderDto import DeliveryOrderDto


class DeliveryCouponBuilder(DataBuilder):

    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        pos_id = int(kwargs["posId"])
        order_id = int(kwargs["orderId"])
        delivery_order_dto = self._get_delivery_order_dto(
            pos_id=pos_id,
            order_id=order_id,
        )

        self._get_header_data(
            data_to_build=data_to_build,
            delivery_order_dto=delivery_order_dto,
        )
        self._get_body_data(
            data_to_build=data_to_build,
            delivery_order_dto=delivery_order_dto,
        )
        self._get_footer_data(data_to_build=data_to_build)

        font_size_small = self.constants.get("DELIVERY_REPORT_FONT_SIZE", "small").lower() == "small"
        data_to_build["DELIVERY_REPORT_FONT_SIZE_SMALL"] = font_size_small
        data_to_build["DELIVERY_REPORT_FONT_SIZE_LARGE"] = not font_size_small

        return data_to_build

    def _get_body_data(
        self,
        data_to_build: Dict[str, JSON],
        delivery_order_dto: DeliveryOrderDto,
    ) -> None:

        self._get_customer_info_data(
            data_to_build=data_to_build,
            delivery_order_dto=delivery_order_dto,
            show_customer_phone=self.constants.get("DELIVERY_COUPON_SHOW_CUSTOMER_PHONE", "false").lower() == "true",
        )
        self._get_items_data(
            data_to_build=data_to_build,
            delivery_order_dto=delivery_order_dto,
        )
        self._get_totals_data(
            data_to_build=data_to_build,
            delivery_order_dto=delivery_order_dto,
        )
        self._get_tenders_data(
            data_to_build=data_to_build,
            delivery_order_dto=delivery_order_dto,
        )
        self._get_observations_data(
            data_to_build=data_to_build,
            delivery_order_dto=delivery_order_dto,
        )

    def _get_delivery_order_dto(
        self,
        pos_id: int,
        order_id: int,
    ) -> DeliveryOrderDto:

        order_service = self.order_api_repository.get_order_service(pos_id=pos_id)
        order_picture_xml = order_service.get_order_picture(order_id=order_id)

        delivery_json = order_picture_xml.find(".//OrderProperty[@key='REMOTE_ORDER_JSON']")
        if delivery_json is None:
            delivery_json = order_picture_xml.find(".//OrderProperty[@key='MANUAL_DELIVERY_ORDER_JSON']")
        delivery_json = json.loads(delivery_json.get("value"))

        return DeliveryOrderDto(
            order_picture_xml=order_picture_xml,
            delivery_json=delivery_json,
            product_repository=self.product_repository,
        )

    @staticmethod
    def _get_header_data(
        data_to_build: Dict[str, JSON],
        delivery_order_dto: DeliveryOrderDto,
    ) -> None:

        data_to_build["PARTNER_NAME"] = delivery_order_dto.display_partner.upper()
        data_to_build["SHORT_REFERENCE"] = delivery_order_dto.order_number
        data_to_build["ORDER_ID"] = delivery_order_dto.pos_order_id

        if delivery_order_dto.partner.lower() != "delivery":
            created_at = convert_from_utf_to_localtime(utc=delivery_order_dto.order_date)
        else:
            created_at = delivery_order_dto.order_date

        data_to_build["CREATED_AT"] = created_at.isoformat()

    @staticmethod
    def _get_customer_info_data(
        data_to_build: Dict[str, JSON],
        delivery_order_dto: DeliveryOrderDto,
        show_customer_phone: bool,
    ) -> None:

        data_to_build["CUSTOMER_NAME"] = delivery_order_dto.customer_name
        if show_customer_phone and (delivery_order_dto.partner.lower() != "ifood" or delivery_order_dto.need_logistics):
            data_to_build["PHONE"] = delivery_order_dto.customer_phone

        data_to_build["ADDRESS"] = delivery_order_dto.address
        data_to_build["COMPLEMENT"] = delivery_order_dto.address_complement
        data_to_build["REFERENCE"] = delivery_order_dto.reference
        data_to_build["NEIGHBORHOOD"] = delivery_order_dto.address_neighborhood
        data_to_build["CITY"] = delivery_order_dto.city
        data_to_build["POSTAL_CODE"] = delivery_order_dto.postal_code

        if delivery_order_dto.take_out_order:
            data_to_build["TAKE_OUT_ORDER"] = True

    @staticmethod
    def _get_items_data(
        data_to_build: Dict[str, JSON],
        delivery_order_dto: DeliveryOrderDto,
    ) -> None:

        items = []
        for item in delivery_order_dto.items:
            items.append(
                {
                    "ITEM_PRODUCT_QUANTITY": item.quantity or "",
                    "ITEM_PRODUCT_NAME": item.product_name,
                    "ITEM_PRODUCT_VALUE": item.price or "",
                }
            )

        data_to_build["ITEM_PRODUCT"] = items

    @staticmethod
    def _get_totals_data(
        data_to_build: Dict[str, JSON],
        delivery_order_dto: DeliveryOrderDto,
    ) -> None:

        data_to_build["SUB_TOTAL_PRICE"] = delivery_order_dto.sub_total_price
        if not delivery_order_dto.take_out_order:
            data_to_build["DELIVERY_FEE"] = delivery_order_dto.delivery_fee
        data_to_build["ADDITIONAL_FEE"] = delivery_order_dto.additional_fee
        data_to_build["TOTAL_PRICE"] = delivery_order_dto.total_price

        vouchers = []
        for key, value in delivery_order_dto.vouchers.items():
            if float(value) > 0:
                vouchers.append({
                    "VOUCHER_NAME": key,
                    "VOUCHER_VALUE": value,
                })

        data_to_build["VOUCHER"] = vouchers or None

    @staticmethod
    def _get_tenders_data(
        data_to_build: Dict[str, JSON],
        delivery_order_dto: DeliveryOrderDto,
    ) -> None:

        tenders = []
        for tender in delivery_order_dto.tenders:
            tenders.append(
                {
                    "TENDER_DESCRIPTION": tender.tender_type,
                    "TENDER_VALUE": tender.value,
                }
            )

            if tender.tender_type_description:
                tenders.append(
                    {
                        "TENDER_DESCRIPTION": tender.tender_type_description,
                        "TENDER_VALUE": tender.tender_change_amount or "",
                    }
                )

        data_to_build["TENDER"] = tenders

    @staticmethod
    def _get_observations_data(
        data_to_build: Dict[str, JSON],
        delivery_order_dto: DeliveryOrderDto,
    ) -> None:

        observations = []
        if delivery_order_dto.observation:
            for observation in delivery_order_dto.observation.split(";"):
                observations.append(
                    {
                        "OBSERVATION_DESCRIPTION": observation.strip(),
                    }
                )

        data_to_build["OBSERVATION"] = observations or None

        if delivery_order_dto.customer_doc:
            data_to_build["CUSTOMER_DOC"] = format_customer_doc(document=delivery_order_dto.customer_doc)

        data_to_build["ORDER_OBSERVATION"] = delivery_order_dto.observations or None

    def _get_footer_data(
        self,
        data_to_build: Dict[str, JSON],
    ) -> None:

        image_path = self.images_path["3S_CHECKOUT_LOGO_FOOTER"]
        if not image_path:
            return

        data_to_build["STORE_IMG_LOGO"] = get_logo_image_str(
            image_path=image_path,
            base_width=250,
        )



from typing import (
    Dict,
    List,
    Optional,
)

from nreports.repository import (
    I18NRepository,
    L10NRepository,
)
from nreports_api.shared import JSON
from ._Line import Line


class Report(object):

    def __init__(
        self,
        *args: List[Dict[str, JSON]],
    ) -> None:

        self.lines: List[Line] = []
        for line in args[0]:
            response = Line.build(line=line)
            if isinstance(response, list):
                self.lines.extend(response)
            else:
                self.lines.append(response)

    def get_data_to_build(
        self,
    ) -> Dict[str, Optional[Dict[str, None]]]:

        ret = {}
        for line in self.lines:
            for name, items in line.get_data_to_build().items():
                ret[name] = items

        return ret

    def format_data(
        self,
        built_data: Dict[str, JSON],
        i18n_repository: I18NRepository,
        l10n_repository: L10NRepository,
    ) -> List[Dict[str, str]]:

        ret = []
        for line in self.lines:
            ret.extend(
                line.format(
                    data=built_data,
                    i18n_repository=i18n_repository,
                    l10n_repository=l10n_repository,
                )
            )

        return ret



from messagebus import (
    MessageBus,
    Message,
    DataType,
)

from .dto import (
    PosUpdaterDto,
    BinariesUpdaterDto,
    ChromeUpdaterDto,
)
from .token import PosUpdaterApiTokens


class PosUpdaterService(object):

    def __init__(
        self,           # type: PosUpdaterService
        message_bus,    # type: MessageBus
        pos_id,         # type: int
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.pos_id = pos_id

        self.component_name = "PosUpdater"

    def catalog_update(
        self,       # type: PosUpdaterService
        user_id,    # type: str
    ):
        # type: (...) -> Message

        data = self._get_common_update_data(user_id=user_id)
        timeout = 10 * 1000 * 1000
        return self._send_message(
            token=PosUpdaterApiTokens.TK_POS_UPDATER_PERFORM_CATALOG_UPDATE,
            timeout=timeout,
            data=str(data),
        )

    def binaries_update(
        self,           # type: PosUpdaterService
        user_id,        # type: str
        force_update,   # type: bool
    ):
        # type: (...) -> Message

        data = BinariesUpdaterDto(
            pos_id=self.pos_id,
            user_id=user_id,
            force_update=force_update,
        )
        return self._send_message(
            token=PosUpdaterApiTokens.TK_POS_UPDATER_PERFORM_BINARIES_UPDATE,
            timeout=60000000 * 20,
            data=str(data),
        )

    def chrome_update(
        self,                # type: PosUpdaterService
        user_id,             # type: str
        validate_version,    # type: bool
    ):
        # type: (...) -> Message

        data = ChromeUpdaterDto(
            pos_id=self.pos_id,
            user_id=user_id,
            validate_version=validate_version,
        )
        return self._send_message(
            token=PosUpdaterApiTokens.TK_POS_UPDATER_PERFORM_CHROME_UPDATE,
            timeout=60000000 * 20,
            data=str(data),
        )

    def loader_update(
        self,       # type: PosUpdaterService
        user_id,    # type: str
    ):
        # type: (...) -> Message

        data = self._get_common_update_data(user_id=user_id)
        return self._send_message(
            token=PosUpdaterApiTokens.TK_POS_UPDATER_PERFORM_LOADER_UPDATE,
            timeout=60000000 * 20,
            data=str(data),
        )

    def _send_message(
        self,       # type: PosUpdaterService
        token,      # type: int
        timeout,    # type: int
        data,       # type: str
    ):
        # type: (...) -> Message

        return self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=token,
                data_type=DataType.json,
                timeout=timeout,
                data=data,
            ),
        )

    def _get_common_update_data(
        self,       # type: PosUpdaterService
        user_id,    # type: str
    ):
        # type: (...) -> PosUpdaterDto

        return PosUpdaterDto(
            pos_id=self.pos_id,
            user_id=user_id,
        )



# Copyright 2022 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common types for gRPC Sync API"""

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from grpc._cython import cygrpc

if TYPE_CHECKING:
    from grpc import ServicerContext
    from grpc._server import _RPCState

RequestType = TypeVar("RequestType")
ResponseType = TypeVar("ResponseType")
SerializingFunction = Callable[[Any], bytes]
DeserializingFunction = Callable[[bytes], Any]
MetadataType = Sequence[Tuple[str, Union[str, bytes]]]
ChannelArgumentType = Tuple[str, Any]
DoneCallbackType = Callable[[Any], None]
NullaryCallbackType = Callable[[], None]
RequestIterableType = Iterable[Any]
ResponseIterableType = Iterable[Any]
UserTag = Callable[[cygrpc.BaseEvent], bool]
IntegratedCallFactory = Callable[
    [
        int,
        bytes,
        None,
        Optional[float],
        Optional[MetadataType],
        Optional[cygrpc.CallCredentials],
        Sequence[Sequence[cygrpc.Operation]],
        UserTag,
        Any,
    ],
    cygrpc.IntegratedCall,
]
ServerTagCallbackType = Tuple[
    Optional["_RPCState"], Sequence[NullaryCallbackType]
]
ServerCallbackTag = Callable[[cygrpc.BaseEvent], ServerTagCallbackType]
ArityAgnosticMethodHandler = Union[
    Callable[
        [RequestType, "ServicerContext", Callable[[ResponseType], None]],
        ResponseType,
    ],
    Callable[
        [RequestType, "ServicerContext", Callable[[ResponseType], None]],
        Iterator[ResponseType],
    ],
    Callable[
        [
            Iterator[RequestType],
            "ServicerContext",
            Callable[[ResponseType], None],
        ],
        ResponseType,
    ],
    Callable[
        [
            Iterator[RequestType],
            "ServicerContext",
            Callable[[ResponseType], None],
        ],
        Iterator[ResponseType],
    ],
    Callable[[RequestType, "ServicerContext"], ResponseType],
    Callable[[RequestType, "ServicerContext"], Iterator[ResponseType]],
    Callable[[Iterator[RequestType], "ServicerContext"], ResponseType],
    Callable[
        [Iterator[RequestType], "ServicerContext"], Iterator[ResponseType]
    ],
]



from ._TableError import TableError
from typing import Dict


class TabToJoinWithPartialTenders(TableError):

    def __init__(
        self,       # type: TabToJoinWithPartialTenders
        table_id,   # type: str
    ):
        # type: (...) -> None

        message = "$TAB_TO_JOIN_WITH_PARTIAL_TENDERS|{}".format(table_id)
        super(TabToJoinWithPartialTenders, self).__init__(message)

        self.table_id = table_id

    def get_extra_data(
        self,   # type: TabToJoinWithPartialTenders
    ):
        # type: (...) -> Dict[str, str]

        return {
            "tableId": self.table_id,
        }



# -*- coding: utf-8 -*-

import sysactions
from bustoken import TK_REMOTE_ORDER_VOID_REMOTE_ORDER
from msgbus import TK_SYS_ACK
from systools import sys_log_exception

from .. import mb_context


def void_remote_order(
    pos_id,                     # type: str
    remote_id,                  # type: str
):
    # type: (...) -> bool

    try:
        token = TK_REMOTE_ORDER_VOID_REMOTE_ORDER
        data = str(remote_id)
        timeout = 10000 * 1000
        msg = mb_context.MB_EasySendMessage('RemoteOrder', token, data=data, timeout=timeout)

        if msg.token == TK_SYS_ACK:
            return True

        model = sysactions.get_model(pos_id)
        error_message = sysactions.translate_message(model, "UNKNOWN_ERROR")
        if msg.data:
            error_message = sysactions.translate_message(
                model=model,
                message_key=msg.data.replace("$", ""),
            )

        sysactions.show_messagebox(
            posid=pos_id,
            message=error_message,
        )
        return False

    except Exception:
        sys_log_exception("Error voiding delivery order")
        raise



# isrc.py - functions for International Standard Recording Codes (ISRC)
# coding: utf-8
#
# Copyright (C) 2021 Nuno André Novo
# Copyright (C) 2021 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ISRC (International Standard Recording Code).

The ISRC is an international standard code (ISO 3901) for uniquely
identifying sound recordings and music video recordings.

More information:

* https://en.wikipedia.org/wiki/International_Standard_Recording_Code

>>> validate('US-SKG-19-12345')
'USSKG1912345'
>>> validate('XX-SKG-19-12345')
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""

import re

from stdnum.exceptions import *
from stdnum.isin import _iso_3116_1_country_codes
from stdnum.util import clean


# An ISRC is composed of a country code, a registrant code a year of
# reference and designation code.
_isrc_re = re.compile(
    r'^(?P<country>[A-Z]{2})(?P<registrant>[A-Z0-9]{3})(?P<year>[0-9]{2})(?P<record>[0-9]{5})$')


# These special codes are allowed for ISRC
_country_codes = set(_iso_3116_1_country_codes + [
    'QM',  # US new registrants due to US codes became exhausted
    'CP',  # reserved for further overflow
    'DG',  # reserved for further overflow
    'ZZ',  # International ISRC Agency codes
])


def compact(number):
    """Convert the ISRC to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip().upper()


def validate(number):
    """Check if the number provided is a valid ISRC. This checks the length,
    the alphabet, and the country code but does not check if the registrant
    code is known."""
    number = compact(number)
    if len(number) != 12:
        raise InvalidLength()
    match = _isrc_re.search(number)
    if not match:
        raise InvalidFormat()
    if match.group('country') not in _country_codes:
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number provided is a valid ISRC."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number, separator='-'):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return separator.join((number[0:2], number[2:5], number[5:7], number[7:]))



#
# distutils/version.py
#
# Implements multiple version numbering conventions for the
# Python Module Distribution Utilities.
#
# $Id$
#

"""Provides classes to represent module version numbers (one class for
each style of version numbering).  There are currently two such classes
implemented: StrictVersion and LooseVersion.

Every version number class implements the following interface:
  * the 'parse' method takes a string and parses it to some internal
    representation; if the string is an invalid version number,
    'parse' raises a ValueError exception
  * the class constructor takes an optional string argument which,
    if supplied, is passed to 'parse'
  * __str__ reconstructs the string that was passed to 'parse' (or
    an equivalent string -- ie. one that will generate an equivalent
    version number instance)
  * __repr__ generates Python code to recreate the version number instance
  * _cmp compares the current instance with either another instance
    of the same class or a string (which will be parsed to an instance
    of the same class, thus must follow the same rules)
"""

import contextlib
import re
import warnings


@contextlib.contextmanager
def suppress_known_deprecation():
    with warnings.catch_warnings(record=True) as ctx:
        warnings.filterwarnings(
            action='default',
            category=DeprecationWarning,
            message="distutils Version classes are deprecated.",
        )
        yield ctx


class Version:
    """Abstract base class for version numbering classes.  Just provides
    constructor (__init__) and reproducer (__repr__), because those
    seem to be the same for all version numbering classes; and route
    rich comparisons to _cmp.
    """

    def __init__(self, vstring=None):
        if vstring:
            self.parse(vstring)
        warnings.warn(
            "distutils Version classes are deprecated. "
            "Use packaging.version instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    def __repr__(self):
        return f"{self.__class__.__name__} ('{self}')"

    def __eq__(self, other):
        c = self._cmp(other)
        if c is NotImplemented:
            return c
        return c == 0

    def __lt__(self, other):
        c = self._cmp(other)
        if c is NotImplemented:
            return c
        return c < 0

    def __le__(self, other):
        c = self._cmp(other)
        if c is NotImplemented:
            return c
        return c <= 0

    def __gt__(self, other):
        c = self._cmp(other)
        if c is NotImplemented:
            return c
        return c > 0

    def __ge__(self, other):
        c = self._cmp(other)
        if c is NotImplemented:
            return c
        return c >= 0


# Interface for version-number classes -- must be implemented
# by the following classes (the concrete ones -- Version should
# be treated as an abstract class).
#    __init__ (string) - create and take same action as 'parse'
#                        (string parameter is optional)
#    parse (string)    - convert a string representation to whatever
#                        internal representation is appropriate for
#                        this style of version numbering
#    __str__ (self)    - convert back to a string; should be very similar
#                        (if not identical to) the string supplied to parse
#    __repr__ (self)   - generate Python code to recreate
#                        the instance
#    _cmp (self, other) - compare two version numbers ('other' may
#                        be an unparsed version string, or another
#                        instance of your version class)


class StrictVersion(Version):
    """Version numbering for anal retentives and software idealists.
    Implements the standard interface for version number classes as
    described above.  A version number consists of two or three
    dot-separated numeric components, with an optional "pre-release" tag
    on the end.  The pre-release tag consists of the letter 'a' or 'b'
    followed by a number.  If the numeric components of two version
    numbers are equal, then one with a pre-release tag will always
    be deemed earlier (lesser) than one without.

    The following are valid version numbers (shown in the order that
    would be obtained by sorting according to the supplied cmp function):

        0.4       0.4.0  (these two are equivalent)
        0.4.1
        0.5a1
        0.5b3
        0.5
        0.9.6
        1.0
        1.0.4a3
        1.0.4b1
        1.0.4

    The following are examples of invalid version numbers:

        1
        2.7.2.2
        1.3.a4
        1.3pl1
        1.3c4

    The rationale for this version numbering system will be explained
    in the distutils documentation.
    """

    version_re = re.compile(
        r'^(\d+) \. (\d+) (\. (\d+))? ([ab](\d+))?$', re.VERBOSE | re.ASCII
    )

    def parse(self, vstring):
        match = self.version_re.match(vstring)
        if not match:
            raise ValueError(f"invalid version number '{vstring}'")

        (major, minor, patch, prerelease, prerelease_num) = match.group(1, 2, 4, 5, 6)

        if patch:
            self.version = tuple(map(int, [major, minor, patch]))
        else:
            self.version = tuple(map(int, [major, minor])) + (0,)

        if prerelease:
            self.prerelease = (prerelease[0], int(prerelease_num))
        else:
            self.prerelease = None

    def __str__(self):
        if self.version[2] == 0:
            vstring = '.'.join(map(str, self.version[0:2]))
        else:
            vstring = '.'.join(map(str, self.version))

        if self.prerelease:
            vstring = vstring + self.prerelease[0] + str(self.prerelease[1])

        return vstring

    def _cmp(self, other):
        if isinstance(other, str):
            with suppress_known_deprecation():
                other = StrictVersion(other)
        elif not isinstance(other, StrictVersion):
            return NotImplemented

        if self.version == other.version:
            # versions match; pre-release drives the comparison
            return self._cmp_prerelease(other)

        return -1 if self.version < other.version else 1

    def _cmp_prerelease(self, other):
        """
        case 1: self has prerelease, other doesn't; other is greater
        case 2: self doesn't have prerelease, other does: self is greater
        case 3: both or neither have prerelease: compare them!
        """
        if self.prerelease and not other.prerelease:
            return -1
        elif not self.prerelease and other.prerelease:
            return 1

        if self.prerelease == other.prerelease:
            return 0
        elif self.prerelease < other.prerelease:
            return -1
        else:
            return 1


# end class StrictVersion


# The rules according to Greg Stein:
# 1) a version number has 1 or more numbers separated by a period or by
#    sequences of letters. If only periods, then these are compared
#    left-to-right to determine an ordering.
# 2) sequences of letters are part of the tuple for comparison and are
#    compared lexicographically
# 3) recognize the numeric components may have leading zeroes
#
# The LooseVersion class below implements these rules: a version number
# string is split up into a tuple of integer and string components, and
# comparison is a simple tuple comparison.  This means that version
# numbers behave in a predictable and obvious way, but a way that might
# not necessarily be how people *want* version numbers to behave.  There
# wouldn't be a problem if people could stick to purely numeric version
# numbers: just split on period and compare the numbers as tuples.
# However, people insist on putting letters into their version numbers;
# the most common purpose seems to be:
#   - indicating a "pre-release" version
#     ('alpha', 'beta', 'a', 'b', 'pre', 'p')
#   - indicating a post-release patch ('p', 'pl', 'patch')
# but of course this can't cover all version number schemes, and there's
# no way to know what a programmer means without asking him.
#
# The problem is what to do with letters (and other non-numeric
# characters) in a version number.  The current implementation does the
# obvious and predictable thing: keep them as strings and compare
# lexically within a tuple comparison.  This has the desired effect if
# an appended letter sequence implies something "post-release":
# eg. "0.99" < "0.99pl14" < "1.0", and "5.001" < "5.001m" < "5.002".
#
# However, if letters in a version number imply a pre-release version,
# the "obvious" thing isn't correct.  Eg. you would expect that
# "1.5.1" < "1.5.2a2" < "1.5.2", but under the tuple/lexical comparison
# implemented here, this just isn't so.
#
# Two possible solutions come to mind.  The first is to tie the
# comparison algorithm to a particular set of semantic rules, as has
# been done in the StrictVersion class above.  This works great as long
# as everyone can go along with bondage and discipline.  Hopefully a
# (large) subset of Python module programmers will agree that the
# particular flavour of bondage and discipline provided by StrictVersion
# provides enough benefit to be worth using, and will submit their
# version numbering scheme to its domination.  The free-thinking
# anarchists in the lot will never give in, though, and something needs
# to be done to accommodate them.
#
# Perhaps a "moderately strict" version class could be implemented that
# lets almost anything slide (syntactically), and makes some heuristic
# assumptions about non-digits in version number strings.  This could
# sink into special-case-hell, though; if I was as talented and
# idiosyncratic as Larry Wall, I'd go ahead and implement a class that
# somehow knows that "1.2.1" < "1.2.2a2" < "1.2.2" < "1.2.2pl3", and is
# just as happy dealing with things like "2g6" and "1.13++".  I don't
# think I'm smart enough to do it right though.
#
# In any case, I've coded the test suite for this module (see
# ../test/test_version.py) specifically to fail on things like comparing
# "1.2a2" and "1.2".  That's not because the *code* is doing anything
# wrong, it's because the simple, obvious design doesn't match my
# complicated, hairy expectations for real-world version numbers.  It
# would be a snap to fix the test suite to say, "Yep, LooseVersion does
# the Right Thing" (ie. the code matches the conception).  But I'd rather
# have a conception that matches common notions about version numbers.


class LooseVersion(Version):
    """Version numbering for anarchists and software realists.
    Implements the standard interface for version number classes as
    described above.  A version number consists of a series of numbers,
    separated by either periods or strings of letters.  When comparing
    version numbers, the numeric components will be compared
    numerically, and the alphabetic components lexically.  The following
    are all valid version numbers, in no particular order:

        1.5.1
        1.5.2b2
        161
        3.10a
        8.02
        3.4j
        1996.07.12
        3.2.pl0
        3.1.1.6
        2g6
        11g
        0.960923
        2.2beta29
        1.13++
        5.5.kw
        2.0b1pl0

    In fact, there is no such thing as an invalid version number under
    this scheme; the rules for comparison are simple and predictable,
    but may not always give the results you want (for some definition
    of "want").
    """

    component_re = re.compile(r'(\d+ | [a-z]+ | \.)', re.VERBOSE)

    def parse(self, vstring):
        # I've given up on thinking I can reconstruct the version string
        # from the parsed tuple -- so I just store the string here for
        # use by __str__
        self.vstring = vstring
        components = [x for x in self.component_re.split(vstring) if x and x != '.']
        for i, obj in enumerate(components):
            try:
                components[i] = int(obj)
            except ValueError:
                pass

        self.version = components

    def __str__(self):
        return self.vstring

    def __repr__(self):
        return f"LooseVersion ('{self}')"

    def _cmp(self, other):
        if isinstance(other, str):
            other = LooseVersion(other)
        elif not isinstance(other, LooseVersion):
            return NotImplemented

        if self.version == other.version:
            return 0
        if self.version < other.version:
            return -1
        if self.version > other.version:
            return 1


# end class LooseVersion



"""
Python Script Wrapper for Windows
=================================

setuptools includes wrappers for Python scripts that allows them to be
executed like regular windows programs.  There are 2 wrappers, one
for command-line programs, cli.exe, and one for graphical programs,
gui.exe.  These programs are almost identical, function pretty much
the same way, and are generated from the same source file.  The
wrapper programs are used by copying them to the directory containing
the script they are to wrap and with the same name as the script they
are to wrap.
"""

import pathlib
import platform
import subprocess
import sys
import textwrap

import pytest

import pkg_resources
from setuptools.command.easy_install import nt_quote_arg

pytestmark = pytest.mark.skipif(sys.platform != 'win32', reason="Windows only")


class WrapperTester:
    @classmethod
    def prep_script(cls, template):
        python_exe = nt_quote_arg(sys.executable)
        return template % locals()

    @classmethod
    def create_script(cls, tmpdir):
        """
        Create a simple script, foo-script.py

        Note that the script starts with a Unix-style '#!' line saying which
        Python executable to run.  The wrapper will use this line to find the
        correct Python executable.
        """

        script = cls.prep_script(cls.script_tmpl)

        with (tmpdir / cls.script_name).open('w') as f:
            f.write(script)

        # also copy cli.exe to the sample directory
        with (tmpdir / cls.wrapper_name).open('wb') as f:
            w = pkg_resources.resource_string('setuptools', cls.wrapper_source)
            f.write(w)


def win_launcher_exe(prefix):
    """A simple routine to select launcher script based on platform."""
    assert prefix in ('cli', 'gui')
    if platform.machine() == "ARM64":
        return "{}-arm64.exe".format(prefix)
    else:
        return "{}-32.exe".format(prefix)


class TestCLI(WrapperTester):
    script_name = 'foo-script.py'
    wrapper_name = 'foo.exe'
    wrapper_source = win_launcher_exe('cli')

    script_tmpl = textwrap.dedent(
        """
        #!%(python_exe)s
        import sys
        input = repr(sys.stdin.read())
        print(sys.argv[0][-14:])
        print(sys.argv[1:])
        print(input)
        if __debug__:
            print('non-optimized')
        """
    ).lstrip()

    def test_basic(self, tmpdir):
        """
        When the copy of cli.exe, foo.exe in this example, runs, it examines
        the path name it was run with and computes a Python script path name
        by removing the '.exe' suffix and adding the '-script.py' suffix. (For
        GUI programs, the suffix '-script.pyw' is added.)  This is why we
        named out script the way we did.  Now we can run out script by running
        the wrapper:

        This example was a little pathological in that it exercised windows
        (MS C runtime) quoting rules:

        - Strings containing spaces are surrounded by double quotes.

        - Double quotes in strings need to be escaped by preceding them with
          back slashes.

        - One or more backslashes preceding double quotes need to be escaped
          by preceding each of them with back slashes.
        """
        self.create_script(tmpdir)
        cmd = [
            str(tmpdir / 'foo.exe'),
            'arg1',
            'arg 2',
            'arg "2\\"',
            'arg 4\\',
            'arg5 a\\\\b',
        ]
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            text=True,
            encoding="utf-8",
        )
        stdout, stderr = proc.communicate('hello\nworld\n')
        actual = stdout.replace('\r\n', '\n')
        expected = textwrap.dedent(
            r"""
            \foo-script.py
            ['arg1', 'arg 2', 'arg "2\\"', 'arg 4\\', 'arg5 a\\\\b']
            'hello\nworld\n'
            non-optimized
            """
        ).lstrip()
        assert actual == expected

    def test_symlink(self, tmpdir):
        """
        Ensure that symlink for the foo.exe is working correctly.
        """
        script_dir = tmpdir / "script_dir"
        script_dir.mkdir()
        self.create_script(script_dir)
        symlink = pathlib.Path(tmpdir / "foo.exe")
        symlink.symlink_to(script_dir / "foo.exe")

        cmd = [
            str(tmpdir / 'foo.exe'),
            'arg1',
            'arg 2',
            'arg "2\\"',
            'arg 4\\',
            'arg5 a\\\\b',
        ]
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            text=True,
            encoding="utf-8",
        )
        stdout, stderr = proc.communicate('hello\nworld\n')
        actual = stdout.replace('\r\n', '\n')
        expected = textwrap.dedent(
            r"""
            \foo-script.py
            ['arg1', 'arg 2', 'arg "2\\"', 'arg 4\\', 'arg5 a\\\\b']
            'hello\nworld\n'
            non-optimized
            """
        ).lstrip()
        assert actual == expected

    def test_with_options(self, tmpdir):
        """
        Specifying Python Command-line Options
        --------------------------------------

        You can specify a single argument on the '#!' line.  This can be used
        to specify Python options like -O, to run in optimized mode or -i
        to start the interactive interpreter.  You can combine multiple
        options as usual. For example, to run in optimized mode and
        enter the interpreter after running the script, you could use -Oi:
        """
        self.create_script(tmpdir)
        tmpl = textwrap.dedent(
            """
            #!%(python_exe)s  -Oi
            import sys
            input = repr(sys.stdin.read())
            print(sys.argv[0][-14:])
            print(sys.argv[1:])
            print(input)
            if __debug__:
                print('non-optimized')
            sys.ps1 = '---'
            """
        ).lstrip()
        with (tmpdir / 'foo-script.py').open('w') as f:
            f.write(self.prep_script(tmpl))
        cmd = [str(tmpdir / 'foo.exe')]
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            encoding="utf-8",
        )
        stdout, stderr = proc.communicate()
        actual = stdout.replace('\r\n', '\n')
        expected = textwrap.dedent(
            r"""
            \foo-script.py
            []
            ''
            ---
            """
        ).lstrip()
        assert actual == expected


class TestGUI(WrapperTester):
    """
    Testing the GUI Version
    -----------------------
    """

    script_name = 'bar-script.pyw'
    wrapper_source = win_launcher_exe('gui')
    wrapper_name = 'bar.exe'

    script_tmpl = textwrap.dedent(
        """
        #!%(python_exe)s
        import sys
        f = open(sys.argv[1], 'wb')
        bytes_written = f.write(repr(sys.argv[2]).encode('utf-8'))
        f.close()
        """
    ).strip()

    def test_basic(self, tmpdir):
        """Test the GUI version with the simple script, bar-script.py"""
        self.create_script(tmpdir)

        cmd = [
            str(tmpdir / 'bar.exe'),
            str(tmpdir / 'test_output.txt'),
            'Test Argument',
        ]
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            encoding="utf-8",
        )
        stdout, stderr = proc.communicate()
        assert not stdout
        assert not stderr
        with (tmpdir / 'test_output.txt').open('rb') as f_out:
            actual = f_out.read().decode('ascii')
        assert actual == repr('Test Argument')



# -*- coding: utf-8 -*-

from sysactions import (
    action,
    show_messagebox
)
from msgbus import TK_SYS_NAK

from api.pos_updater import get_pos_updater_service
from ..util.get_menu_user_logged import get_menu_manager_authenticated_user


@action
def update_chrome(
    pos_id   # type: str
):
    # type: (...) -> None

    resp = show_messagebox(
        posid=pos_id,
        message="$UPDATE_CHROME_MESSAGE_QUESTION",
        buttons="$OK|$CANCEL"
    )
    if resp == 1:
        return

    user_id = get_menu_manager_authenticated_user(pos_id=int(pos_id))
    updater_service = get_pos_updater_service(pos_id=pos_id)
    reply = updater_service.chrome_update(
        user_id=user_id,
        validate_version=False,
    )

    if reply.token == TK_SYS_NAK or reply.data:
        show_messagebox(
            posid=pos_id,
            message="$UPDATE_CHROME_" + reply.data.upper(),
            buttons="OK",
        )
        return

    show_messagebox(
        posid=pos_id,
        message="$UPDATE_CHROME_MESSAGE",
        buttons="OK",
    )



from msgbus import MBMessage


class UnexpectedResponse(Exception):
    def __init__(
        self,     # type: UnexpectedResponse
        service,  # type: str
        message,  # type: MBMessage
    ):
        # type: (...) -> None

        self.service = service
        self.message = message

    def __str__(self):
        # type: () -> str

        return "Unexpected response from {}: {} - {}".format(self.service, self.message.token, self.message.data)

    def __repr__(self):
        # type: () -> str

        return self.__str__()



import json
import logging
import re

from application.model import ApplierSubTypes
from application.repository.api import APIRepository
from order_api import OrderService
from order_api.dto import CustomProperty
from table_manager_api import TableService
from table_manager_api.dto import (
    TableServiceDto,
    TableStatus,
)
from typing import (
    Dict,
    List,
    Optional,
    Union,
)

from .dto import GetLoyaltyCustomerInfoRequest

logger = logging.getLogger(__name__)


class GetLoyaltyCustomerInfoInteractor(object):

    def __init__(self, api_repository):
        # type: (APIRepository) -> None

        self.api_repository = api_repository

    def get_loyalty_customer_info(
        self,       # type: GetLoyaltyCustomerInfoInteractor
        request,    # type: GetLoyaltyCustomerInfoRequest
    ):
        # type: (...) -> Optional[str]

        logger.info("Requesting customer loyaltySystemId")
        loyalty_info = self.api_repository.retrieve_loyalty_customer_info(loyalty_customer_id=request.loyalty_id)
        logger.info("Retrieved customer loyalty data: {}".format(loyalty_info))
        loyalty_info = self._update_applier_type(customer_info=loyalty_info)
        if loyalty_info:
            self._save_loyalty_info(
                loyalty_info=loyalty_info,
                request=request,
            )
        return json.dumps(loyalty_info)

    @staticmethod
    def _update_applier_type(
        customer_info,  # type: str
    ):
        # type: (...) -> Dict[str, Union[str, Dict, List]]

        if not customer_info:
            return customer_info

        try:
            info = json.loads(customer_info)
            benefits = info.get("benefits", [])
            for benefit in benefits:
                benefit["applier"] = "Loyalty"
                benefit["applierSubType"] = ApplierSubTypes.LOYALTY

            return info
        except ValueError:
            logger.error("Invalid JSON returned: {}".format(customer_info))
            return {}

    def _save_loyalty_info(
        self,           # type: GetLoyaltyCustomerInfoInteractor
        loyalty_info,   # type: Dict[str, Union[str, Dict, List]]
        request,        # type: GetLoyaltyCustomerInfoRequest
    ):
        # type: (...) -> None

        order_service = OrderService(
            pos_id=request.pos_id,
            message_bus=self.api_repository.message_bus,
        )
        table_service = TableService(
            pos_id=request.pos_id,
            message_bus=self.api_repository.message_bus,
        )
        loyalty_properties = self._parse_loyalty_customer_info(
            loyalty_customer_info=loyalty_info,
            request=request,
        )
        if request.table_id:
            return self._save_table_loyalty_info(
                table_service=table_service,
                table_id=request.table_id,
                seat=request.seat,
                loyalty_properties=loyalty_properties,
                order_service=order_service,
            )

        return self._save_order_loyalty_info(
            order_id=request.order_id,
            loyalty_properties=loyalty_properties,
            order_service=order_service,
        )

    def _save_table_loyalty_info(
        self,                   # type: GetLoyaltyCustomerInfoInteractor
        table_service,          # type: TableService
        table_id,               # type: str
        seat,                   # type: str
        loyalty_properties,     # type: Dict[str, Union[str, List, Dict]]
        order_service,          # type: OrderService
    ):
        # type: (...) -> None

        table_service_dto = table_service.get_service(table_id=table_id)
        if table_service_dto.status == TableStatus.Totaled:
            self._save_order_loyalty_info(
                order_id=table_service_dto.payment_order.order_id,
                loyalty_properties=loyalty_properties,
                order_service=order_service,
            )

        table_loyalty_properties = self._format_to_table_loyalty_properties(
            table_service_dto=table_service_dto,
            loyalty_properties=loyalty_properties,
            seat=seat,
        )

        table_service.set_custom_properties(
            properties=table_loyalty_properties,
            table_id=table_id,
        )

    def _format_to_table_loyalty_properties(
        self,                   # type: GetLoyaltyCustomerInfoInteractor
        table_service_dto,      # type: TableServiceDto
        loyalty_properties,     # type: Dict[str, Union[str, List, Dict]]
        seat,                   # type: str
    ):
        # type: (...) -> Dict[str, str]

        table_loyalty_properties = dict(loyalty_properties)

        original_table_loyalty_properties = self._get_table_loyalty_properties(table_service_dto=table_service_dto)
        for key, value in table_loyalty_properties.items():
            if key in ["CUSTOMER_NAME", "CUSTOMER_DOC"]:
                if seat == "table":
                    original_table_loyalty_properties[key] = value
                continue

            if key in original_table_loyalty_properties:
                original_table_loyalty_properties[key].update({seat: value})
            else:
                original_table_loyalty_properties[key] = {seat: value}

            original_table_loyalty_properties[key] = json.dumps(original_table_loyalty_properties[key])

        return original_table_loyalty_properties

    @staticmethod
    def _get_table_loyalty_properties(
        table_service_dto,  # type: TableServiceDto
    ):
        # type: (...) -> Dict[str, Dict]

        return {
            key: json.loads(value)
            for key, value in table_service_dto.custom_properties.items()
            if key in [
                "LOYALTY_CUSTOMER_INFO",
                "LOYALTY_CUSTOMER_BENEFITS",
                "LOYALTY_COUPONS_TO_APPLY_ON_TOTAL",
                "LOYALTY_ID",
            ]
        }

    @staticmethod
    def _save_order_loyalty_info(
        order_id,               # type: int
        loyalty_properties,     # type: Dict[str, Union[str, List, Dict]]
        order_service,          # type: OrderService
    ):
        # type: (...) -> None

        json_fields = ["LOYALTY_CUSTOMER_INFO", "LOYALTY_CUSTOMER_BENEFITS", "LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"]
        order_properties = [
            CustomProperty(
                key=key,
                value=json.dumps(value) if key in json_fields else value,
            ) for key, value in loyalty_properties.items()
        ]

        order_service.set_order_custom_properties(
            custom_properties=order_properties,
            order_id=order_id,
        )

    def _parse_loyalty_customer_info(
        self,                   # type: GetLoyaltyCustomerInfoInteractor
        loyalty_customer_info,  # type: Dict[str, Union[str, Dict, List]]
        request,                # type: GetLoyaltyCustomerInfoRequest
    ):
        # type: (...) -> Dict[str, Union[str, Dict, List]]

        customer_benefits = loyalty_customer_info.pop("benefits", [])
        customer = loyalty_customer_info.get("customerInfo", {}).get("customer", {})
        customer_document = customer.get("document") or ""
        customer["isValidDocument"] = self._validate_document(customer_doc=customer_document) is not False
        loyalty_customer_info.get("customerInfo", {}).update(customer)

        result = {
            "LOYALTY_ID": request.loyalty_id,
            "LOYALTY_CUSTOMER_INFO": loyalty_customer_info,
        }
        if customer.get("name"):
            result["CUSTOMER_NAME"] = customer["name"].encode("utf-8")
        if customer_document:
            result["CUSTOMER_DOC"] = customer_document

        if not customer_benefits:
            return result

        result["LOYALTY_CUSTOMER_BENEFITS"] = customer_benefits

        result["LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"] = [
            benefit.get("id") for benefit in customer_benefits
            if benefit.get("automatic") == "ON_TOTAL"
        ]
        return result

    def _validate_document(
        self,           # type: GetLoyaltyCustomerInfoInteractor
        customer_doc,   # type: str
    ):
        # type: (...) -> bool

        if len(customer_doc) <= 11:
            return self.validate_cpf(cpf=customer_doc)

        return self.validate_cnpj(cnpj=customer_doc)

    @staticmethod
    def validate_cpf(
        cpf,    # type: str
    ):
        # type: (...) -> Union[bool, str]

        cpf = "".join(re.findall(r"\d", str(cpf)))
        if not cpf or len(cpf) < 11:
            return False

        integers = list(map(int, cpf))
        new = integers[:9]

        if all(num == integers[0] for num in integers):
            return False

        while len(new) < 11:
            r = sum([(len(new) + 1 - i) * v for i, v in enumerate(new)]) % 11

            if r > 1:
                f = 11 - r
            else:
                f = 0
            new.append(f)

        if new == integers:
            return cpf

        return False

    @staticmethod
    def validate_cnpj(
        cnpj,   # type: str
    ):
        # type: (...) -> Union[bool, str]

        cnpj = "".join(re.findall(r"\d", str(cnpj)))
        if not cnpj or len(cnpj) < 14:
            return False

        integers = list(map(int, cnpj))
        new = integers[:12]

        if all(num == integers[0] for num in integers):
            return False

        prod = [5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2]
        while len(new) < 14:
            r = sum([x * y for (x, y) in zip(new, prod)]) % 11
            if r > 1:
                f = 11 - r
            else:
                f = 0
            new.append(f)
            prod.insert(0, 6)

        if new == integers:
            return cnpj

        return False



import json

from typing import Optional

from update_api.dto import ChromeUpdaterDto

from messagebus import Message

from ._BaseMessageProcessor import BaseUpdaterProcessor
from application.interactor import ChromeUpdaterInteractor


class ChromeUpdaterProcessor(BaseUpdaterProcessor):
    def __init__(
        self,
        interactor    # type: ChromeUpdaterInteractor
    ):
        # type: (...) -> None

        super(ChromeUpdaterProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(
        self,       # type: ChromeUpdaterProcessor
        message,    # type: Message
    ):
        # type: (...) -> Optional[ChromeUpdaterDto]

        if message and message.data:
            data = json.loads(message.data)
            return ChromeUpdaterDto(
                pos_id=data.get("pos_id"),
                user_id=data.get("user_id"),
                validate_version=data.get("validate_version"),
            )

        return None

    def call_business(
        self,    # type: ChromeUpdaterProcessor
        data,    # type: Optional[ChromeUpdaterDto]
    ):
        # type: (...) -> Message

        return self.interactor.update_chrome(
            validate=(data and data.validate_version),
            data=data,
        )



import re
from sqlite3 import Connection

from dbapitemplate import DbApiTemplate
from production.model import Course
from typing import (
    Dict,
    List,
    Optional,
)

from production.repository import ProductRepository as ProductionProductRepository


class ProductRepository(ProductionProductRepository):

    def __init__(
        self,           # type: ProductRepository
        connection,     # type: Connection
    ):
        # type: (...) -> None

        self.connection = DbApiTemplate(connection)
        self.all_part_code_jit_lines = None

    def get_part_codes_of_jit_lines(
        self,       # type: ProductRepository
        jit_lines,  # type: List[str]
    ):
        # type: (...) -> List[int]

        if len(jit_lines) == 0:
            return []

        part_codes = []
        all_products_jit_lines = self.get_all_part_code_jit_lines()
        for key, value in all_products_jit_lines.items():
            current_jit_lines = re.findall('(\\w+)', value)
            if any(x in jit_lines for x in current_jit_lines):
                part_codes.append(key)

        return part_codes

    def get_all_part_code_jit_lines(
        self,   # type: ProductRepository
    ):
        # type: (...) -> Dict[int, str]

        products_jit_lines = {}

        def mapper(row):
            products_jit_lines[int(row["ProductCode"])] = row["JITLines"].encode("utf-8")

        if self.all_part_code_jit_lines is None:
            self.connection.execute(
                sql=get_part_codes_of_jit_lines_query,
                mapper=mapper,
            )
            self.all_part_code_jit_lines = products_jit_lines

        return self.all_part_code_jit_lines

    def get_items_to_show_on_jit(
        self,               # type: ProductRepository
        allowed_jit_lines,  # type: List
    ):
        # type: (...) -> List[int]

        if not allowed_jit_lines:
            return []

        part_codes = []

        def mapper(row):
            part_codes.append(int(row["ProductCode"]))

        formatted_jit_lines = ["JITLines LIKE '%{}%'".format(s) for s in allowed_jit_lines]
        placeholders = " OR ".join(formatted_jit_lines)
        query = get_show_on_kitchen_by_jit.format(placeholders=placeholders)

        self.connection.execute(
            sql=query,
            mapper=mapper,
        )
        return part_codes

    def get_prep_times(
        self,    # type: ProductRepository
    ):
        # type: (...) -> Optional[Dict[int, int]]

        prep_time_dict = {}

        def mapper(row):
            part_code = int(row["ProductCode"])
            prep_time = int(row["CustomParamValue"])
            prep_time_dict[part_code] = prep_time

        self.connection.execute(
            sql=get_prep_time_query,
            mapper=mapper,
        )
        return prep_time_dict

    def get_courses_products(
        self,   # type: ProductRepository
    ):
        # type: (...) -> Dict[int, List[int]]

        course_dict = {}

        def mapper(row):
            part_code = int(row["ProductCode"])
            course = row["CustomParamValue"].encode("utf-8")
            if course not in course_dict:
                course_dict[course] = []

            course_dict[course].append(part_code)

        self.connection.execute(
            sql=course_query,
            mapper=mapper,
        )
        return course_dict

    def get_not_show_on_kitchen(
        self,   # type: ProductRepository
    ):
        # type: (...) -> List[int]

        part_codes = []

        def mapper(row):
            part_codes.append(int(row["ProductCode"]))

        self.connection.execute(
            sql=not_show_on_kitchen_query,
            mapper=mapper,
        )
        return part_codes

    def get_no_jit_part_codes(
        self,   # type: ProductRepository
    ):
        # type: (...) -> List[int]

        part_codes = []

        def mapper(row):
            part_codes.append(int(row["ProductCode"]))

        self.connection.execute(
            sql=no_jit_query,
            mapper=mapper,
        )
        return part_codes

    def get_cfh_items(
        self,   # type: ProductRepository
    ):
        # type: (...) -> List[int]

        part_codes = []

        def mapper(row):
            part_codes.append(int(row["ProductCode"]))

        self.connection.execute(
            sql=cfh_items_query,
            mapper=mapper,
        )
        return part_codes

    def get_all_courses(
        self,   # type: ProductRepository
    ):
        # type: (...) -> List[Course]

        courses = []

        def mapper(row):
            course = Course(
                course_id=int(row["Id"]),
                name=row["Name"].encode("utf-8"),
                description=row["Description"].encode("utf-8"),
                abbreviation=row["Abbreviation"].encode("utf-8"),
                delay=int(row["Delay"]),
            )
            courses.append(course)

        self.connection.execute(
            sql=get_courses_query,
            mapper=mapper,
        )
        return courses

    def get_products_with_custom_params(
        self,               # type: ProductRepository
        custom_params,      # type: List[str]
    ):
        # type: (...) -> Dict[str, Dict[int, str]]

        product_custom_params = {}

        def mapper(row):
            product_code = int(row["ProductCode"])
            custom_param_id = row["CustomParamId"]
            if custom_param_id not in product_custom_params:
                product_custom_params[custom_param_id] = {}

            product_custom_params[custom_param_id][product_code] = (row["CustomParamValue"] or "").encode("utf-8")

        custom_params = ', '.join('"{0}"'.format(w) for w in custom_params)
        self.connection.execute(
            sql=get_products_with_custom_params.format(customParams=custom_params),
            mapper=mapper,
        )
        return product_custom_params


get_part_codes_of_jit_lines_query = """
SELECT ProductCode, JITLines
FROM
Production
WHERE (Expiration IS NULL OR Expiration > DATETIME('now')) AND JITLInes IS NOT NULL
"""

get_show_on_kitchen_by_jit = """
SELECT *
FROM Production
WHERE ({placeholders})
AND ShowOnKitchen = 1
"""

get_prep_time_query = """
SELECT ProductCode, CustomParamValue
FROM ProductCustomParams
WHERE CustomParamId = 'PrepTime' AND CustomParamValue IS NOT NULL
"""


course_query = """
select ProductCode, CustomParamValue from ProductCustomParams where CustomParamId = 'Course'
"""

not_show_on_kitchen_query = """
SELECT p.ProductCode
FROM Production p
JOIN ProductKernelParams pkp ON pkp.ProductCode = p.ProductCode
JOIN ProductType pt ON pt.TypeId = pkp.ProductType
WHERE ShowOnKitchen == 0 AND pt.TypeDescr <> 'TYPE_OPTION'
"""

no_jit_query = """
select ProductCode from Production where JITLines is NULL and ShowOnKitchen = 1
"""

cfh_items_query = """
 SELECT ProductCode FROM ProductTags
WHERE LOWER(Tag) = 'cfh=true'
UNION
SELECT ProductCode FROM ProductCustomParams
WHERE LOWER(CustomParamId) = 'cfh' AND LOWER(CustomParamValue) = 'true'
"""


get_courses_query = """
SELECT Id, Name, Description, Abbreviation, Delay FROM Courses
"""

get_products_with_custom_params = """
    SELECT
        ProductCode,
        CustomParamId,
        CustomParamValue
    FROM ProductCustomParams
    WHERE CustomParamId IN ({customParams})
"""



from typing import List


class ListenedEvents(object):
    HVAUDIT = "HVAUDIT"
    ORDER_MODIFIED = "ORDER_MODIFIED"
    ORDER_STATE = "ORDER_STATE"
    PRODUCTION_ORDER_EVENT = "PRODUCTION_ORDER_EVENT"
    TABLE_CLOSED = "TABLE_CLOSED"
    ORDER_REPROCESS = "ORDER_REPROCESS"
    POS_STATECHANGED = "POS_STATECHANGED"
    POS_UPDATER_LOGGER = "POS_UPDATER_LOGGER"
    STORE_STATUS = "STORE_STATUS"
    GLOBAL_RESTART = "GLOBAL_RESTART"
    DECLARED_AMOUNT = "DECLARED_AMOUNT"
    TRANSFER_CASH_IN = "TRANSFER_CASH_IN"
    TRANSFER_CASH_OUT = "TRANSFER_CASH_OUT"
    INITIAL_AMOUNT = "INITIAL_AMOUNT"


def all_events():
    # type: () -> List[str]
    return [x for x in vars(ListenedEvents) if not x.startswith("__")]



from sqlalchemy import BigInteger
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy.testing import in_

from ._autogen_fixtures import AutogenFixtureTest
from ... import testing
from ...testing import config
from ...testing import eq_
from ...testing import is_
from ...testing import TestBase


class AlterColumnTest(AutogenFixtureTest, TestBase):
    __backend__ = True

    @testing.combinations((True,), (False,))
    @config.requirements.comments
    def test_all_existings_filled(self, pk):
        m1 = MetaData()
        m2 = MetaData()

        Table("a", m1, Column("x", Integer, primary_key=pk))
        Table("a", m2, Column("x", Integer, comment="x", primary_key=pk))

        alter_col = self._assert_alter_col(m1, m2, pk)
        eq_(alter_col.modify_comment, "x")

    @testing.combinations((True,), (False,))
    @config.requirements.comments
    def test_all_existings_filled_in_notnull(self, pk):
        m1 = MetaData()
        m2 = MetaData()

        Table("a", m1, Column("x", Integer, nullable=False, primary_key=pk))
        Table(
            "a",
            m2,
            Column("x", Integer, nullable=False, comment="x", primary_key=pk),
        )

        self._assert_alter_col(m1, m2, pk, nullable=False)

    @testing.combinations((True,), (False,))
    @config.requirements.comments
    def test_all_existings_filled_in_comment(self, pk):
        m1 = MetaData()
        m2 = MetaData()

        Table("a", m1, Column("x", Integer, comment="old", primary_key=pk))
        Table("a", m2, Column("x", Integer, comment="new", primary_key=pk))

        alter_col = self._assert_alter_col(m1, m2, pk)
        eq_(alter_col.existing_comment, "old")

    @testing.combinations((True,), (False,))
    @config.requirements.comments
    def test_all_existings_filled_in_server_default(self, pk):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "a", m1, Column("x", Integer, server_default="5", primary_key=pk)
        )
        Table(
            "a",
            m2,
            Column(
                "x", Integer, server_default="5", comment="new", primary_key=pk
            ),
        )

        alter_col = self._assert_alter_col(m1, m2, pk)
        in_("5", alter_col.existing_server_default.arg.text)

    def _assert_alter_col(self, m1, m2, pk, nullable=None):
        ops = self._fixture(m1, m2, return_ops=True)
        modify_table = ops.ops[-1]
        alter_col = modify_table.ops[0]

        if nullable is None:
            eq_(alter_col.existing_nullable, not pk)
        else:
            eq_(alter_col.existing_nullable, nullable)
        assert alter_col.existing_type._compare_type_affinity(Integer())
        return alter_col


class AutoincrementTest(AutogenFixtureTest, TestBase):
    __backend__ = True
    __requires__ = ("integer_subtype_comparisons",)

    def test_alter_column_autoincrement_none(self):
        m1 = MetaData()
        m2 = MetaData()

        Table("a", m1, Column("x", Integer, nullable=False))
        Table("a", m2, Column("x", Integer, nullable=True))

        ops = self._fixture(m1, m2, return_ops=True)
        assert "autoincrement" not in ops.ops[0].ops[0].kw

    def test_alter_column_autoincrement_pk_false(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "a",
            m1,
            Column("x", Integer, primary_key=True, autoincrement=False),
        )
        Table(
            "a",
            m2,
            Column("x", BigInteger, primary_key=True, autoincrement=False),
        )

        ops = self._fixture(m1, m2, return_ops=True)
        is_(ops.ops[0].ops[0].kw["autoincrement"], False)

    def test_alter_column_autoincrement_pk_implicit_true(self):
        m1 = MetaData()
        m2 = MetaData()

        Table("a", m1, Column("x", Integer, primary_key=True))
        Table("a", m2, Column("x", BigInteger, primary_key=True))

        ops = self._fixture(m1, m2, return_ops=True)
        is_(ops.ops[0].ops[0].kw["autoincrement"], True)

    def test_alter_column_autoincrement_pk_explicit_true(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "a", m1, Column("x", Integer, primary_key=True, autoincrement=True)
        )
        Table(
            "a",
            m2,
            Column("x", BigInteger, primary_key=True, autoincrement=True),
        )

        ops = self._fixture(m1, m2, return_ops=True)
        is_(ops.ops[0].ops[0].kw["autoincrement"], True)

    def test_alter_column_autoincrement_nonpk_false(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "a",
            m1,
            Column("id", Integer, primary_key=True),
            Column("x", Integer, autoincrement=False),
        )
        Table(
            "a",
            m2,
            Column("id", Integer, primary_key=True),
            Column("x", BigInteger, autoincrement=False),
        )

        ops = self._fixture(m1, m2, return_ops=True)
        is_(ops.ops[0].ops[0].kw["autoincrement"], False)

    def test_alter_column_autoincrement_nonpk_implicit_false(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "a",
            m1,
            Column("id", Integer, primary_key=True),
            Column("x", Integer),
        )
        Table(
            "a",
            m2,
            Column("id", Integer, primary_key=True),
            Column("x", BigInteger),
        )

        ops = self._fixture(m1, m2, return_ops=True)
        assert "autoincrement" not in ops.ops[0].ops[0].kw

    def test_alter_column_autoincrement_nonpk_explicit_true(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "a",
            m1,
            Column("id", Integer, primary_key=True, autoincrement=False),
            Column("x", Integer, autoincrement=True),
        )
        Table(
            "a",
            m2,
            Column("id", Integer, primary_key=True, autoincrement=False),
            Column("x", BigInteger, autoincrement=True),
        )

        ops = self._fixture(m1, m2, return_ops=True)
        is_(ops.ops[0].ops[0].kw["autoincrement"], True)

    def test_alter_column_autoincrement_compositepk_false(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "a",
            m1,
            Column("id", Integer, primary_key=True),
            Column("x", Integer, primary_key=True, autoincrement=False),
        )
        Table(
            "a",
            m2,
            Column("id", Integer, primary_key=True),
            Column("x", BigInteger, primary_key=True, autoincrement=False),
        )

        ops = self._fixture(m1, m2, return_ops=True)
        is_(ops.ops[0].ops[0].kw["autoincrement"], False)

    def test_alter_column_autoincrement_compositepk_implicit_false(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "a",
            m1,
            Column("id", Integer, primary_key=True),
            Column("x", Integer, primary_key=True),
        )
        Table(
            "a",
            m2,
            Column("id", Integer, primary_key=True),
            Column("x", BigInteger, primary_key=True),
        )

        ops = self._fixture(m1, m2, return_ops=True)
        assert "autoincrement" not in ops.ops[0].ops[0].kw

    @config.requirements.autoincrement_on_composite_pk
    def test_alter_column_autoincrement_compositepk_explicit_true(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "a",
            m1,
            Column("id", Integer, primary_key=True, autoincrement=False),
            Column("x", Integer, primary_key=True, autoincrement=True),
            # on SQLA 1.0 and earlier, this being present
            # trips the "add KEY for the primary key" so that the
            # AUTO_INCREMENT keyword is accepted by MySQL.  SQLA 1.1 and
            # greater the columns are just reorganized.
            mysql_engine="InnoDB",
        )
        Table(
            "a",
            m2,
            Column("id", Integer, primary_key=True, autoincrement=False),
            Column("x", BigInteger, primary_key=True, autoincrement=True),
        )

        ops = self._fixture(m1, m2, return_ops=True)
        is_(ops.ops[0].ops[0].kw["autoincrement"], True)



import json
import logging
import xml.etree.cElementTree as eTree
from xml.etree import ElementTree

from discount_api.token import (
    TK_DISCOUNT_GET_JUSTIFICATION_CONFIG,
    TK_DISCOUNT_ON_TABLE_OPEN,
)

from discount_api.dto import JustificationConfigDto
from messagebus import (
    DataType,
    DefaultToken,
    Message,
    MessageBus,
    MessageBusException,
)
from typing import Optional

logger = logging.getLogger(__name__)


class DiscountService(object):
    def __init__(
        self,               # type: DiscountService
        message_bus,        # type: MessageBus
    ):
        self.message_bus = message_bus

    def on_service_open(
        self,               # type: DiscountService
        pos_id,             # type: int
        table_xml,          # type: ElementTree.Element
    ):
        # type: (...) -> Optional[int]

        data = json.dumps(
            dict(
                pos_id=pos_id,
                table_xml=eTree.tostring(
                    element=table_xml,
                    encoding="unicode",
                ),
            )
        ).encode("utf-8")

        try:
            reply = self.message_bus.send_message(
                component="nDiscount",
                message=Message(
                    token=TK_DISCOUNT_ON_TABLE_OPEN,
                    data_type=DataType.param,
                    data=data,
                    timeout=30000000,
                ),
            )

            if reply.token == DefaultToken.TK_SYS_ACK:
                if reply.data == b"":
                    return None

                return int(reply.data)

        except MessageBusException:
            logger.warning("Error applying discount on service open", exc_info=True)

        return None

    def get_justification_config(
        self,               # type: DiscountService
    ):
        # type: (...) -> JustificationConfigDto

        reply = self.message_bus.send_message(
            component="nDiscount",
            message=Message(
                token=TK_DISCOUNT_GET_JUSTIFICATION_CONFIG,
                data_type=DataType.param,
                data=b"",
                timeout=30000000,
            ),
        )

        if reply.token == DefaultToken.TK_SYS_ACK:
            reply_dict = json.loads(reply.data.decode("utf-8"))
            return JustificationConfigDto(
                justifications=reply_dict["justifications"],
            )

        try:
            data_text = reply.data.decode("utf-8")
        except UnicodeDecodeError:
            data_text = reply.data.hex()

        error_message = "Error retrieving justification config. Token: [{}], data: [{}]".format(
            reply.token,
            data_text,
        )

        raise RuntimeError(error_message)



import logging
from typing import Any

from botocore.exceptions import EndpointConnectionError
from messagebus import (
    Event,
    MessageBus,
)

logger = logging.getLogger("TopicClientReceiveThread")


class SqsConsumer(object):
    def __init__(
        self,
        message_bus: MessageBus,
        queue: Any,
        store_code: str,
    ) -> None:

        self.message_bus = message_bus
        self.queue = queue
        self.store_code = store_code
        self.running = True

        logger.info("Starting sqs consumer...")

    def run(self):
        while self.running:
            try:
                messages = self.queue.receive_messages(
                    AttributeNames=['All'],
                    MaxNumberOfMessages=10,
                    MessageAttributeNames=['ToStore', 'Subject'],
                    VisibilityTimeout=5,
                    WaitTimeSeconds=20
                )

                if len(messages) != 0:
                    logger.info("Received messages: {}".format(len(messages)))

                for message in messages:
                    message.delete()

                    try:
                        if message.message_attributes is None:
                            logger.error("Message without attributes: {}".format(message))
                            continue

                        if "ToStore" not in message.message_attributes:
                            logger.error("Message without ToStore: {}".format(message))
                            continue

                        to_store = message.message_attributes["ToStore"]
                        if "StringValue" not in to_store:
                            logger.error("Message without ToStore value: {}".format(message))
                            continue

                        to_store_value = to_store["StringValue"]
                        index = to_store_value.find("_")
                        if index < 0:
                            store_code = to_store_value
                        else:
                            store_code = to_store_value[index + 1:]

                        if store_code != str(self.store_code):
                            logger.error("Message not for this store: {}".format(message))
                            continue

                        if "Subject" not in message.message_attributes:
                            logger.error("Message without subject: {}".format(message))
                            continue

                        if "StringValue" not in message.message_attributes["Subject"]:
                            logger.error("Message without subject value: {}".format(message))
                            continue

                        event = Event(
                            subject=message.message_attributes["Subject"]["StringValue"],
                            event_type="",
                            data=message.body.encode("utf-8"),
                        )
                        logger.info("Publishing event: {} - {}".format(event.subject, event.data))
                        self.message_bus.publish_event(event)
                    except (Exception,):
                        logger.exception("Error publishing event")
            except EndpointConnectionError:
                logger.warning("Error Receiving Messages - Connection Error")
            except (Exception,):
                logger.exception("Error Receiving Messages - Unknown Error")

    def stop(self):
        logger.info("Stopping sqs consumer...")

        self.running = False



import json
from typing import List

from production.box import ViewBox
from production.view import PrinterView


class GetPrintersInfoInteractor(object):
    def __init__(self, view_boxes):
        # type: (List[ViewBox]) -> None
        self.view_boxes = view_boxes

    def execute(self):
        # type: () -> str
        info = []
        for view_box in self.view_boxes:
            view = view_box.view
            if not isinstance(view, PrinterView):
                continue

            printer_name = view.get_printer_name()
            if printer_name is None:
                continue

            if printer_name not in info:
                info.append(printer_name)

        return json.dumps(info)



from typing import (
    Dict,
    List,
)

from ._TableError import TableError


class MissingAnyOfParameters(TableError):

    def __init__(
        self,           # type: MissingAnyOfParameters
        params=None,    # type: List[str]
    ):
        # type: (...) -> None

        self.params = params

    def get_extra_data(
        self,   # type: MissingAnyOfParameters
    ):
        # type: (...) -> Dict[str, List[str]]

        return {
            "params": self.params,
        }



#! python
#
# This module implements a simple socket based client.
# It does not support changing any port parameters and will silently ignore any
# requests to do so.
#
# The purpose of this module is that applications using pySerial can connect to
# TCP/IP to serial port converters that do not support RFC 2217.
#
# This file is part of pySerial. https://github.com/pyserial/pyserial
# (C) 2001-2015 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
#
# URL format:    socket://<host>:<port>[/option[/option...]]
# options:
# - "debug" print diagnostic messages

from __future__ import absolute_import

import errno
import logging
import select
import socket
import time
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from serial.serialutil import SerialBase, SerialException, to_bytes, \
    PortNotOpenError, SerialTimeoutException, Timeout

# map log level names to constants. used in from_url()
LOGGER_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
}

POLL_TIMEOUT = 5


class Serial(SerialBase):
    """Serial port implementation for plain sockets."""

    BAUDRATES = (50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
                 9600, 19200, 38400, 57600, 115200)

    def open(self):
        """\
        Open port with current settings. This may throw a SerialException
        if the port cannot be opened.
        """
        self.logger = None
        if self._port is None:
            raise SerialException("Port must be configured before it can be used.")
        if self.is_open:
            raise SerialException("Port is already open.")
        try:
            # timeout is used for write timeout support :/ and to get an initial connection timeout
            self._socket = socket.create_connection(self.from_url(self.portstr), timeout=POLL_TIMEOUT)
        except Exception as msg:
            self._socket = None
            raise SerialException("Could not open port {}: {}".format(self.portstr, msg))
        # after connecting, switch to non-blocking, we're using select
        self._socket.setblocking(False)

        # not that there is anything to configure...
        self._reconfigure_port()
        # all things set up get, now a clean start
        self.is_open = True
        if not self._dsrdtr:
            self._update_dtr_state()
        if not self._rtscts:
            self._update_rts_state()
        self.reset_input_buffer()
        self.reset_output_buffer()

    def _reconfigure_port(self):
        """\
        Set communication parameters on opened port. For the socket://
        protocol all settings are ignored!
        """
        if self._socket is None:
            raise SerialException("Can only operate on open ports")
        if self.logger:
            self.logger.info('ignored port configuration change')

    def close(self):
        """Close port"""
        if self.is_open:
            if self._socket:
                try:
                    self._socket.shutdown(socket.SHUT_RDWR)
                    self._socket.close()
                except:
                    # ignore errors.
                    pass
                self._socket = None
            self.is_open = False
            # in case of quick reconnects, give the server some time
            time.sleep(0.3)

    def from_url(self, url):
        """extract host and port from an URL string"""
        parts = urlparse.urlsplit(url)
        if parts.scheme != "socket":
            raise SerialException(
                'expected a string in the form '
                '"socket://<host>:<port>[?logging={debug|info|warning|error}]": '
                'not starting with socket:// ({!r})'.format(parts.scheme))
        try:
            # process options now, directly altering self
            for option, values in urlparse.parse_qs(parts.query, True).items():
                if option == 'logging':
                    logging.basicConfig()   # XXX is that good to call it here?
                    self.logger = logging.getLogger('pySerial.socket')
                    self.logger.setLevel(LOGGER_LEVELS[values[0]])
                    self.logger.debug('enabled logging')
                else:
                    raise ValueError('unknown option: {!r}'.format(option))
            if not 0 <= parts.port < 65536:
                raise ValueError("port not in range 0...65535")
        except ValueError as e:
            raise SerialException(
                'expected a string in the form '
                '"socket://<host>:<port>[?logging={debug|info|warning|error}]": {}'.format(e))

        return (parts.hostname, parts.port)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    @property
    def in_waiting(self):
        """Return the number of bytes currently in the input buffer."""
        if not self.is_open:
            raise PortNotOpenError()
        # Poll the socket to see if it is ready for reading.
        # If ready, at least one byte will be to read.
        lr, lw, lx = select.select([self._socket], [], [], 0)
        return len(lr)

    # select based implementation, similar to posix, but only using socket API
    # to be portable, additionally handle socket timeout which is used to
    # emulate write timeouts
    def read(self, size=1):
        """\
        Read size bytes from the serial port. If a timeout is set it may
        return less characters as requested. With no timeout it will block
        until the requested number of bytes is read.
        """
        if not self.is_open:
            raise PortNotOpenError()
        read = bytearray()
        timeout = Timeout(self._timeout)
        while len(read) < size:
            try:
                ready, _, _ = select.select([self._socket], [], [], timeout.time_left())
                # If select was used with a timeout, and the timeout occurs, it
                # returns with empty lists -> thus abort read operation.
                # For timeout == 0 (non-blocking operation) also abort when
                # there is nothing to read.
                if not ready:
                    break   # timeout
                buf = self._socket.recv(size - len(read))
                # read should always return some data as select reported it was
                # ready to read when we get to this point, unless it is EOF
                if not buf:
                    raise SerialException('socket disconnected')
                read.extend(buf)
            except OSError as e:
                # this is for Python 3.x where select.error is a subclass of
                # OSError ignore BlockingIOErrors and EINTR. other errors are shown
                # https://www.python.org/dev/peps/pep-0475.
                if e.errno not in (errno.EAGAIN, errno.EALREADY, errno.EWOULDBLOCK, errno.EINPROGRESS, errno.EINTR):
                    raise SerialException('read failed: {}'.format(e))
            except (select.error, socket.error) as e:
                # this is for Python 2.x
                # ignore BlockingIOErrors and EINTR. all errors are shown
                # see also http://www.python.org/dev/peps/pep-3151/#select
                if e[0] not in (errno.EAGAIN, errno.EALREADY, errno.EWOULDBLOCK, errno.EINPROGRESS, errno.EINTR):
                    raise SerialException('read failed: {}'.format(e))
            if timeout.expired():
                break
        return bytes(read)

    def write(self, data):
        """\
        Output the given byte string over the serial port. Can block if the
        connection is blocked. May raise SerialException if the connection is
        closed.
        """
        if not self.is_open:
            raise PortNotOpenError()

        d = to_bytes(data)
        tx_len = length = len(d)
        timeout = Timeout(self._write_timeout)
        while tx_len > 0:
            try:
                n = self._socket.send(d)
                if timeout.is_non_blocking:
                    # Zero timeout indicates non-blocking - simply return the
                    # number of bytes of data actually written
                    return n
                elif not timeout.is_infinite:
                    # when timeout is set, use select to wait for being ready
                    # with the time left as timeout
                    if timeout.expired():
                        raise SerialTimeoutException('Write timeout')
                    _, ready, _ = select.select([], [self._socket], [], timeout.time_left())
                    if not ready:
                        raise SerialTimeoutException('Write timeout')
                else:
                    assert timeout.time_left() is None
                    # wait for write operation
                    _, ready, _ = select.select([], [self._socket], [], None)
                    if not ready:
                        raise SerialException('write failed (select)')
                d = d[n:]
                tx_len -= n
            except SerialException:
                raise
            except OSError as e:
                # this is for Python 3.x where select.error is a subclass of
                # OSError ignore BlockingIOErrors and EINTR. other errors are shown
                # https://www.python.org/dev/peps/pep-0475.
                if e.errno not in (errno.EAGAIN, errno.EALREADY, errno.EWOULDBLOCK, errno.EINPROGRESS, errno.EINTR):
                    raise SerialException('write failed: {}'.format(e))
            except select.error as e:
                # this is for Python 2.x
                # ignore BlockingIOErrors and EINTR. all errors are shown
                # see also http://www.python.org/dev/peps/pep-3151/#select
                if e[0] not in (errno.EAGAIN, errno.EALREADY, errno.EWOULDBLOCK, errno.EINPROGRESS, errno.EINTR):
                    raise SerialException('write failed: {}'.format(e))
            if not timeout.is_non_blocking and timeout.expired():
                raise SerialTimeoutException('Write timeout')
        return length - len(d)

    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""
        if not self.is_open:
            raise PortNotOpenError()

        # just use recv to remove input, while there is some
        ready = True
        while ready:
            ready, _, _ = select.select([self._socket], [], [], 0)
            try:
                if ready:
                    ready = self._socket.recv(4096)
            except OSError as e:
                # this is for Python 3.x where select.error is a subclass of
                # OSError ignore BlockingIOErrors and EINTR. other errors are shown
                # https://www.python.org/dev/peps/pep-0475.
                if e.errno not in (errno.EAGAIN, errno.EALREADY, errno.EWOULDBLOCK, errno.EINPROGRESS, errno.EINTR):
                    raise SerialException('read failed: {}'.format(e))
            except (select.error, socket.error) as e:
                # this is for Python 2.x
                # ignore BlockingIOErrors and EINTR. all errors are shown
                # see also http://www.python.org/dev/peps/pep-3151/#select
                if e[0] not in (errno.EAGAIN, errno.EALREADY, errno.EWOULDBLOCK, errno.EINPROGRESS, errno.EINTR):
                    raise SerialException('read failed: {}'.format(e))

    def reset_output_buffer(self):
        """\
        Clear output buffer, aborting the current output and
        discarding all that is in the buffer.
        """
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('ignored reset_output_buffer')

    def send_break(self, duration=0.25):
        """\
        Send break condition. Timed, returns to idle state after given
        duration.
        """
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('ignored send_break({!r})'.format(duration))

    def _update_break_state(self):
        """Set break: Controls TXD. When active, to transmitting is
        possible."""
        if self.logger:
            self.logger.info('ignored _update_break_state({!r})'.format(self._break_state))

    def _update_rts_state(self):
        """Set terminal status line: Request To Send"""
        if self.logger:
            self.logger.info('ignored _update_rts_state({!r})'.format(self._rts_state))

    def _update_dtr_state(self):
        """Set terminal status line: Data Terminal Ready"""
        if self.logger:
            self.logger.info('ignored _update_dtr_state({!r})'.format(self._dtr_state))

    @property
    def cts(self):
        """Read terminal status line: Clear To Send"""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('returning dummy for cts')
        return True

    @property
    def dsr(self):
        """Read terminal status line: Data Set Ready"""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('returning dummy for dsr')
        return True

    @property
    def ri(self):
        """Read terminal status line: Ring Indicator"""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('returning dummy for ri')
        return False

    @property
    def cd(self):
        """Read terminal status line: Carrier Detect"""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('returning dummy for cd)')
        return True

    # - - - platform specific - - -

    # works on Linux and probably all the other POSIX systems
    def fileno(self):
        """Get the file handle of the underlying socket for use with select"""
        return self._socket.fileno()


#
# simple client test
if __name__ == '__main__':
    import sys
    s = Serial('socket://localhost:7000')
    sys.stdout.write('{}\n'.format(s))

    sys.stdout.write("write...\n")
    s.write(b"hello\n")
    s.flush()
    sys.stdout.write("read: {}\n".format(s.read(5)))

    s.close()



import json

from messagebus import Message
from typing import Optional
from update_api.dto import PosUpdaterDto

from ._BaseMessageProcessor import BaseUpdaterProcessor
from application.interactor import LoaderUpdaterInteractor
from application.util import settings


class LoaderUpdaterProcessor(BaseUpdaterProcessor):
    def __init__(
        self,           # type: LoaderUpdaterProcessor
        interactor,     # type: LoaderUpdaterInteractor
    ):
        # type: (...) -> None
        super(LoaderUpdaterProcessor, self).__init__()

        self.interactor = interactor

    def parse_data(
        self,       # type: LoaderUpdaterProcessor
        message,    # type: Message
    ):
        # type: (...) -> Optional[PosUpdaterDto]

        if message and message.data:
            data = json.loads(message.data)
            return PosUpdaterDto(
                pos_id=data.get("pos_id"),
                user_id=data.get("user_id"),
            )

        return None

    def call_business(
        self,   # type: LoaderUpdaterProcessor
        data,   # type: Optional[PosUpdaterDto]
    ):
        # type: (...) -> Message

        settings.clear_status()
        return self.interactor.update_loaders(data=data)



from cfgtools import Group
from typing import List
from util import (
    find_key_value,
    find_key_values,
)

from ._AutoOrderCommand import AutoOrderCommand
from ._Statistics import Statistics
from ._Threshold import Threshold
from ._ZoomLevel import ZoomLevel


class KdsView(object):
    def __init__(self, group):
        # type: (Group) -> None

        self.name = ''
        self.production_view = ''
        self.selected = False
        self.selected_order_id = -1
        self.display_mode = ''
        self.title = ''
        self.header_text_macro = ''
        self.body_text_macro = ''
        self.body_sale_line_macro = ''
        self.footer_text_macro = ''
        self.carousel = True
        self.on_double_click = ''
        self.on_sell_click = 'served'
        self.show_items = True
        self.zoom_levels = []
        self.selected_zoom = 0
        self.thresholds = []
        self.statistics = []
        self.auto_order_command = None
        self.colored_area = 'body'
        self.view_with_actions = True
        self.show_in_preparation_items = False
        self.display_type = ''
        self.custom_font_size = None
        self.custom_font_color = None
        self.sound_notification_for_new_orders = None

        self.parse_kds_view(group)

    def parse_kds_view(self, view_group):
        # type: (Group) -> ()
        self.name = view_group.name

        self.production_view = find_key_value(view_group, "ProductionView")
        self.header_text_macro = find_key_value(view_group, "CellHeader")
        self.body_text_macro = find_key_value(view_group, "CellBody")
        default_sale_line_body = "#{itemIcon} #{itemLevel} #{itemQuantity} #{itemModifier} #{itemPrefix} #{itemName}"
        self.body_sale_line_macro = find_key_value(view_group, "SaleLineBody", default_sale_line_body)
        self.footer_text_macro = find_key_value(view_group, "CellFooter")
        self.on_double_click = find_key_value(view_group, "CellDoubleClick")
        self.on_sell_click = find_key_value(view_group, "OnSellClick")

        self.carousel = find_key_value(view_group, "Carousel", "true").lower() == 'true'
        self.show_items = find_key_value(view_group, "ShowItems", "true").lower() == 'true'

        for group in view_group.groups:
            if group.name == 'ZoomLevels':
                self.zoom_levels = self._get_zoom_levels(group)

            if group.name == 'Statistics':
                self.statistics = self._set_statistics(group)

            if group.name == 'Customization':
                self._set_custom_parameters(customization_group=group)

        for value in find_key_values(view_group, 'Thresholds'):
            self.thresholds.append(Threshold(value))

        auto_order_command_config = find_key_value(view_group, "AutoOrderCommand")
        if auto_order_command_config:
            self.auto_order_command = AutoOrderCommand(auto_order_command_config)

        colored_area_config = find_key_value(view_group, "ColoredArea")
        if colored_area_config:
            self.colored_area = colored_area_config

        self.view_with_actions = find_key_value(view_group, "ViewWithActions", "true").lower() == 'true'

    @staticmethod
    def _get_zoom_levels(group):
        zoom_levels = []
        for zoom_level_group in group.groups:
            zoom_levels.append(ZoomLevel(zoom_level_group))

        return zoom_levels

    def _set_statistics(self, group):
        for statistic_group in group.groups:
            name = statistic_group.name
            variable = statistic_group.find_value("variable")
            statistics_format = statistic_group.find_value("format")
            goal = statistic_group.find_value("goal.value")
            comparison = statistic_group.find_value("goal.comparison")

            statistic = Statistics(name, variable, statistics_format, goal, comparison)
            self.statistics.append(statistic)

        return self.statistics

    def _set_custom_parameters(
        self,                       # type: KdsView
        customization_group,        # type: List[Group]
    ):
        self.custom_font_size = find_key_value(
            view_group=customization_group,
            key_name="CustomFontSize",
            default_value=None,
        )
        self.custom_font_color = find_key_value(
            view_group=customization_group,
            key_name="CustomFontColor",
            default_value=None,
        )

        sound_notification_for_new_orders = find_key_value(
            view_group=customization_group,
            key_name="soundNotificationForNewOrders",
            default_value=False,
        )
        self.sound_notification_for_new_orders = str(sound_notification_for_new_orders.lower() == 'true')



import json

from typing import Optional


class SkimType(object):
    ALL = 0
    SKIM = 1
    CHANGE_FUND = 2


class GLAccount(object):

    def __init__(
        self,               # type: GLAccount
        manager="",         # type: str
        envelope="",        # type: str
        justification="",   # type: str
        skim_type=None,     # type: Optional[int]
    ):
        # type: (...) -> None

        self.manager = manager
        self.envelope = envelope
        self.justification = justification
        if skim_type is not None:
            self.type = skim_type

    def to_json(
        self,   # type: GLAccount
    ):
        # type: (...) -> str

        return json.dumps(
            obj=self,
            default=lambda o: o.__dict__,
            sort_keys=True,
        )



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from html.parser import HTMLParser
from itertools import zip_longest

PRIORITY_PARENT_TAGS = ('code', 'a')
OMIT_NESTED_TAGS = ('span', 'i', 'code', 'a')
OMIT_SELF_TAGS = ('i', 'b')
HTML_BLOCK_DISPLAY_TAGS = ('p', 'note', 'ul', 'li')


class DocStringParser(HTMLParser):
    """
    A simple HTML parser.  Focused on converting the subset of HTML
    that appears in the documentation strings of the JSON models into
    simple ReST format.
    """

    def __init__(self, doc):
        self.tree = None
        self.doc = doc
        super().__init__()

    def reset(self):
        HTMLParser.reset(self)
        self.tree = HTMLTree(self.doc)

    def feed(self, data):
        super().feed(data)
        self.tree.write()
        self.tree = HTMLTree(self.doc)

    def close(self):
        super().close()
        # Write if there is anything remaining.
        self.tree.write()
        self.tree = HTMLTree(self.doc)

    def handle_starttag(self, tag, attrs):
        self.tree.add_tag(tag, attrs=attrs)

    def handle_endtag(self, tag):
        self.tree.add_tag(tag, is_start=False)

    def handle_data(self, data):
        self.tree.add_data(data)


class HTMLTree:
    """
    A tree which handles HTML nodes. Designed to work with a python HTML parser,
    meaning that the current_node will be the most recently opened tag. When
    a tag is closed, the current_node moves up to the parent node.
    """

    def __init__(self, doc):
        self.doc = doc
        self.head = StemNode()
        self.current_node = self.head
        self.unhandled_tags = []

    def add_tag(self, tag, attrs=None, is_start=True):
        if not self._doc_has_handler(tag, is_start):
            self.unhandled_tags.append(tag)
            return

        if is_start:
            node = TagNode(tag, attrs)
            self.current_node.add_child(node)
            self.current_node = node
        else:
            self.current_node = self.current_node.parent

    def _doc_has_handler(self, tag, is_start):
        if is_start:
            handler_name = 'start_%s' % tag
        else:
            handler_name = 'end_%s' % tag

        return hasattr(self.doc.style, handler_name)

    def add_data(self, data):
        self.current_node.add_child(DataNode(data))

    def write(self):
        self.head.write(self.doc)


class Node:
    def __init__(self, parent=None):
        self.parent = parent

    def write(self, doc):
        raise NotImplementedError


class StemNode(Node):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.children = []

    def add_child(self, child):
        child.parent = self
        self.children.append(child)

    def write(self, doc):
        self.collapse_whitespace()
        self._write_children(doc)

    def _write_children(self, doc):
        for child, next_child in zip_longest(self.children, self.children[1:]):
            if isinstance(child, TagNode) and next_child is not None:
                child.write(doc, next_child)
            else:
                child.write(doc)

    def is_whitespace(self):
        return all(child.is_whitespace() for child in self.children)

    def startswith_whitespace(self):
        return self.children and self.children[0].startswith_whitespace()

    def endswith_whitespace(self):
        return self.children and self.children[-1].endswith_whitespace()

    def lstrip(self):
        while self.children and self.children[0].is_whitespace():
            self.children = self.children[1:]
        if self.children:
            self.children[0].lstrip()

    def rstrip(self):
        while self.children and self.children[-1].is_whitespace():
            self.children = self.children[:-1]
        if self.children:
            self.children[-1].rstrip()

    def collapse_whitespace(self):
        """Remove collapsible white-space from HTML.

        HTML in docstrings often contains extraneous white-space around tags,
        for readability. Browsers would collapse this white-space before
        rendering. If not removed before conversion to RST where white-space is
        part of the syntax, for example for indentation, it can result in
        incorrect output.
        """
        self.lstrip()
        self.rstrip()
        for child in self.children:
            child.collapse_whitespace()


class TagNode(StemNode):
    """
    A generic Tag node. It will verify that handlers exist before writing.
    """

    def __init__(self, tag, attrs=None, parent=None):
        super().__init__(parent)
        self.attrs = attrs
        self.tag = tag

    def _has_nested_tags(self):
        # Returns True if any children are TagNodes and False otherwise.
        return any(isinstance(child, TagNode) for child in self.children)

    def write(self, doc, next_child=None):
        prioritize_nested_tags = (
            self.tag in OMIT_SELF_TAGS and self._has_nested_tags()
        )
        prioritize_parent_tag = (
            isinstance(self.parent, TagNode)
            and self.parent.tag in PRIORITY_PARENT_TAGS
            and self.tag in OMIT_NESTED_TAGS
        )
        if prioritize_nested_tags or prioritize_parent_tag:
            self._write_children(doc)
            return

        self._write_start(doc)
        self._write_children(doc)
        self._write_end(doc, next_child)

    def collapse_whitespace(self):
        """Remove collapsible white-space.

        All tags collapse internal whitespace. Block-display HTML tags also
        strip all leading and trailing whitespace.

        Approximately follows the specification used in browsers:
        https://www.w3.org/TR/css-text-3/#white-space-rules
        https://developer.mozilla.org/en-US/docs/Web/API/Document_Object_Model/Whitespace
        """
        if self.tag in HTML_BLOCK_DISPLAY_TAGS:
            self.lstrip()
            self.rstrip()
        # Collapse whitespace in situations like ``</b> <i> foo</i>`` into
        # ``</b><i> foo</i>``.
        for prev, cur in zip(self.children[:-1], self.children[1:]):
            if (
                isinstance(prev, DataNode)
                and prev.endswith_whitespace()
                and cur.startswith_whitespace()
            ):
                cur.lstrip()
        # Same logic, but for situations like ``<b>bar </b> <i>``:
        for cur, nxt in zip(self.children[:-1], self.children[1:]):
            if (
                isinstance(nxt, DataNode)
                and cur.endswith_whitespace()
                and nxt.startswith_whitespace()
            ):
                cur.rstrip()
        # Recurse into children
        for child in self.children:
            child.collapse_whitespace()

    def _write_start(self, doc):
        handler_name = 'start_%s' % self.tag
        if hasattr(doc.style, handler_name):
            getattr(doc.style, handler_name)(self.attrs)

    def _write_end(self, doc, next_child):
        handler_name = 'end_%s' % self.tag
        if hasattr(doc.style, handler_name):
            if handler_name == 'end_a':
                # We use lookahead to determine if a space is needed after a link node
                getattr(doc.style, handler_name)(next_child)
            else:
                getattr(doc.style, handler_name)()


class DataNode(Node):
    """
    A Node that contains only string data.
    """

    def __init__(self, data, parent=None):
        super().__init__(parent)
        if not isinstance(data, str):
            raise ValueError("Expecting string type, %s given." % type(data))
        self._leading_whitespace = ''
        self._trailing_whitespace = ''
        self._stripped_data = ''
        if data == '':
            return
        if data.isspace():
            self._trailing_whitespace = data
            return
        first_non_space = next(
            idx for idx, ch in enumerate(data) if not ch.isspace()
        )
        last_non_space = len(data) - next(
            idx for idx, ch in enumerate(reversed(data)) if not ch.isspace()
        )
        self._leading_whitespace = data[:first_non_space]
        self._trailing_whitespace = data[last_non_space:]
        self._stripped_data = data[first_non_space:last_non_space]

    @property
    def data(self):
        return (
            f'{self._leading_whitespace}{self._stripped_data}'
            f'{self._trailing_whitespace}'
        )

    def is_whitespace(self):
        return self._stripped_data == '' and (
            self._leading_whitespace != '' or self._trailing_whitespace != ''
        )

    def startswith_whitespace(self):
        return self._leading_whitespace != '' or (
            self._stripped_data == '' and self._trailing_whitespace != ''
        )

    def endswith_whitespace(self):
        return self._trailing_whitespace != '' or (
            self._stripped_data == '' and self._leading_whitespace != ''
        )

    def lstrip(self):
        if self._leading_whitespace != '':
            self._leading_whitespace = ''
        elif self._stripped_data == '':
            self.rstrip()

    def rstrip(self):
        if self._trailing_whitespace != '':
            self._trailing_whitespace = ''
        elif self._stripped_data == '':
            self.lstrip()

    def collapse_whitespace(self):
        """Noop, ``DataNode.write`` always collapses whitespace"""
        return

    def write(self, doc):
        words = doc.translate_words(self._stripped_data.split())
        str_data = (
            f'{self._leading_whitespace}{" ".join(words)}'
            f'{self._trailing_whitespace}'
        )
        if str_data != '':
            doc.handle_data(str_data)



from abc import ABCMeta, abstractmethod

from production.model import ProductionOrder, ProdStates
from typing import Union, List
from ._ProductionBox import ProductionBox


def _set_invalid_if_no_items(order):
    if not order.items:
        order.prod_state = ProdStates.INVALID


class OrderChangerProductionBox(ProductionBox):
    __metaclass__ = ABCMeta

    def order_modified(self, order):
        if not order.state:
            self.debug("{} - Order without state".format(self.name))

        initial_order_prod_state = order.prod_state
        if initial_order_prod_state == ProdStates.INVALID:
            self.debug("{:50} - Order ProdState [INVALID] - No modifications...", self.name)
        else:
            self.debug("{:50} - Order before modification:\n{}", self.name, str(order))
        altered_order = self.change_order(order)
        if isinstance(altered_order, list) and len(altered_order) > 0:
            for index, order_temp in enumerate(altered_order):
                try:
                    begin_message = "#region    - Order ID: {} - Order Index {}"
                    self.debug(begin_message, order_temp.order_id, index, extra={'simple': True})
                    self.debug("{:50} - Order Index {} after modification:\n{}", self.name, index, str(order_temp))
                    self.send_to_children(order_temp)
                finally:
                    final_message = "#endregion - Order ID: {} - Order Index {}"
                    self.debug(final_message, order_temp.order_id, index, extra={'simple': True})

        elif altered_order:
            _set_invalid_if_no_items(altered_order)
            if initial_order_prod_state != ProdStates.INVALID:
                self.debug("{:50} - Order after modification:\n{}", self.name, str(altered_order))
            self.send_to_children(altered_order)

        else:
            self.debug("{:50} - Order removed", self.name)

    @abstractmethod
    def change_order(self, order):
        # type: (ProductionOrder) -> Union[ProductionOrder, List[ProductionOrder]]

        raise NotImplementedError()



from setuptools import setup, find_packages

setup(
    name="PosUpdaterApi",
    version="1.0.0",
    packages=find_packages(where="."),
    install_requires=["MessageBus>=1.1.0"],
    url="https://www.e-deploy.com.br",
    author="Claudinei Nascimento",
    author_email="claudinei.nascimento@e-deploy.com.br",
)



import sqlalchemy as sa
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table

from alembic.util import sqla_compat
from ._autogen_fixtures import AutogenFixtureTest
from ... import testing
from ...testing import config
from ...testing import eq_
from ...testing import is_true
from ...testing import TestBase


class AutogenerateIdentityTest(AutogenFixtureTest, TestBase):
    __requires__ = ("identity_columns",)
    __backend__ = True

    def test_add_identity_column(self):
        m1 = MetaData()
        m2 = MetaData()

        Table("user", m1, Column("other", sa.Text))

        Table(
            "user",
            m2,
            Column("other", sa.Text),
            Column(
                "id",
                Integer,
                sa.Identity(start=5, increment=7),
                primary_key=True,
            ),
        )

        diffs = self._fixture(m1, m2)

        eq_(diffs[0][0], "add_column")
        eq_(diffs[0][2], "user")
        eq_(diffs[0][3].name, "id")
        i = diffs[0][3].identity

        is_true(isinstance(i, sa.Identity))
        eq_(i.start, 5)
        eq_(i.increment, 7)

    def test_remove_identity_column(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "user",
            m1,
            Column(
                "id",
                Integer,
                sa.Identity(start=2, increment=3),
                primary_key=True,
            ),
        )

        Table("user", m2)

        diffs = self._fixture(m1, m2)

        eq_(diffs[0][0], "remove_column")
        eq_(diffs[0][2], "user")
        c = diffs[0][3]
        eq_(c.name, "id")

        is_true(isinstance(c.identity, sa.Identity))
        eq_(c.identity.start, 2)
        eq_(c.identity.increment, 3)

    def test_no_change_identity_column(self):
        m1 = MetaData()
        m2 = MetaData()

        for m in (m1, m2):
            id_ = sa.Identity(start=2)
            Table("user", m, Column("id", Integer, id_))

        diffs = self._fixture(m1, m2)

        eq_(diffs, [])

    def test_dialect_kwargs_changes(self):
        m1 = MetaData()
        m2 = MetaData()

        if sqla_compat.identity_has_dialect_kwargs:
            args = {"oracle_on_null": True, "oracle_order": True}
        else:
            args = {"on_null": True, "order": True}

        Table("user", m1, Column("id", Integer, sa.Identity(start=2)))
        id_ = sa.Identity(start=2, **args)
        Table("user", m2, Column("id", Integer, id_))

        diffs = self._fixture(m1, m2)
        if config.db.name == "oracle":
            is_true(len(diffs), 1)
            eq_(diffs[0][0][0], "modify_default")
        else:
            eq_(diffs, [])

    @testing.combinations(
        (None, dict(start=2)),
        (dict(start=2), None),
        (dict(start=2), dict(start=2, increment=7)),
        (dict(always=False), dict(always=True)),
        (
            dict(start=1, minvalue=0, maxvalue=100, cycle=True),
            dict(start=1, minvalue=0, maxvalue=100, cycle=False),
        ),
        (
            dict(start=10, increment=3, maxvalue=9999),
            dict(start=10, increment=1, maxvalue=3333),
        ),
    )
    @config.requirements.identity_columns_alter
    def test_change_identity(self, before, after):
        arg_before = (sa.Identity(**before),) if before else ()
        arg_after = (sa.Identity(**after),) if after else ()

        m1 = MetaData()
        m2 = MetaData()

        Table(
            "user",
            m1,
            Column("id", Integer, *arg_before),
            Column("other", sa.Text),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, *arg_after),
            Column("other", sa.Text),
        )

        diffs = self._fixture(m1, m2)

        eq_(len(diffs[0]), 1)
        diffs = diffs[0][0]
        eq_(diffs[0], "modify_default")
        eq_(diffs[2], "user")
        eq_(diffs[3], "id")
        old = diffs[5]
        new = diffs[6]

        def check(kw, idt):
            if kw:
                is_true(isinstance(idt, sa.Identity))
                for k, v in kw.items():
                    eq_(getattr(idt, k), v)
            else:
                is_true(idt in (None, False))

        check(before, old)
        check(after, new)

    def test_add_identity_to_column(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "user",
            m1,
            Column("id", Integer),
            Column("other", sa.Text),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, sa.Identity(start=2, maxvalue=1000)),
            Column("other", sa.Text),
        )

        diffs = self._fixture(m1, m2)

        eq_(len(diffs[0]), 1)
        diffs = diffs[0][0]
        eq_(diffs[0], "modify_default")
        eq_(diffs[2], "user")
        eq_(diffs[3], "id")
        eq_(diffs[5], None)
        added = diffs[6]

        is_true(isinstance(added, sa.Identity))
        eq_(added.start, 2)
        eq_(added.maxvalue, 1000)

    def test_remove_identity_from_column(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "user",
            m1,
            Column("id", Integer, sa.Identity(start=2, maxvalue=1000)),
            Column("other", sa.Text),
        )

        Table(
            "user",
            m2,
            Column("id", Integer),
            Column("other", sa.Text),
        )

        diffs = self._fixture(m1, m2)

        eq_(len(diffs[0]), 1)
        diffs = diffs[0][0]
        eq_(diffs[0], "modify_default")
        eq_(diffs[2], "user")
        eq_(diffs[3], "id")
        eq_(diffs[6], None)
        removed = diffs[5]

        is_true(isinstance(removed, sa.Identity))



class GetLoyaltyCustomerInfoRequest(object):

    def __init__(
        self,
        pos_id,         # type: int
        loyalty_id,     # type: str
        order_id,       # type: int
        table_id,       # type: str
        seat,           # type: str
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.loyalty_id = loyalty_id
        self.order_id = order_id
        self.table_id = table_id
        self.seat = seat
